import math
from erdb.typing.params import ParamRow
from erdb.typing.enums import SpEffectType, AttackType
from erdb.typing.effects import AttributeField, EffectModel


_TRIGGER_FIELDS = [
    "stateInfo", "invocationConditionsStateChange1",
    "invocationConditionsStateChange2", "invocationConditionsStateChange3"
]

_CONDITION_FIELDS = [
    "magicSubCategoryChange1", "magicSubCategoryChange2", "magicSubCategoryChange3"
]

_PVE_TO_PVP = {
    "defEnemyDmgCorrectRate_Physics": "defPlayerDmgCorrectRate_Physics",
    "defEnemyDmgCorrectRate_Magic":   "defPlayerDmgCorrectRate_Magic",
    "defEnemyDmgCorrectRate_Fire":    "defPlayerDmgCorrectRate_Fire",
    "defEnemyDmgCorrectRate_Thunder": "defPlayerDmgCorrectRate_Thunder",
    "defEnemyDmgCorrectRate_Dark":    "defPlayerDmgCorrectRate_Dark",
    "atkEnemyDmgCorrectRate_Physics": "atkPlayerDmgCorrectRate_Physics",
    "atkEnemyDmgCorrectRate_Magic":   "atkPlayerDmgCorrectRate_Magic",
    "atkEnemyDmgCorrectRate_Fire":    "atkPlayerDmgCorrectRate_Fire",
    "atkEnemyDmgCorrectRate_Thunder": "atkPlayerDmgCorrectRate_Thunder",
    "atkEnemyDmgCorrectRate_Dark":    "atkPlayerDmgCorrectRate_Dark",
}

def _floor_decimal_2(value: float) -> float:
    value = round(value, 6) # do not floor cases like 1.89999999999
    return math.floor(value * 100) / 100.0

def conditions(sp_effect: ParamRow, triggeree: ParamRow | None = None) -> list[str] | None:
    conds = set()

    def _append_triggers(source: ParamRow):
        for trigger in _TRIGGER_FIELDS:
            effect_type = SpEffectType(source[trigger])

            if not effect_type.is_passive():
                conds.add(str(effect_type))

            if effect_type == SpEffectType.SPELL_POWER_BOOST:
                boostSorcery = ("magParamChange", "Affects Sorceries")
                boostIncantation = ("miracleParamChange", "Affects Incantations")
                for field, cond in [boostSorcery, boostIncantation]:
                    if sp_effect[field].as_bool:
                        conds.add(cond)

    def _append_conditions(source: ParamRow):
        for cond in _CONDITION_FIELDS:
            attack_type_str = source[cond]
            if (cond := AttackType(attack_type_str)) != AttackType.NONE:
                conds.add(str(cond))

    for field, direction in [("conditionHp", "below"), ("conditionHpRate", "above")]:
        if cond := sp_effect[field].get_int():
            conds.add(f"HP {direction} {cond}%")

    _append_triggers(sp_effect)
    _append_conditions(sp_effect)

    if triggeree:
        _append_triggers(triggeree)
        _append_conditions(triggeree)

    return None if len(conds) == 0 else sorted(list(conds))

def interval(sp_effect: ParamRow) -> float | None:
    interv = sp_effect["motionInterval"]
    return None if interv == "0" else float(interv)

def value_pvp(sp_effect: ParamRow, field_pve: str, attrib_fields: dict[str, AttributeField]) -> float | None:
    if not (field_pvp := _PVE_TO_PVP.get(field_pve, None)):
        return None

    attrib_field = attrib_fields[field_pve]
    val = sp_effect[field_pvp].as_float

    return attrib_field.parser(val, attrib_field.effect_model)

def generic(value: float, model: EffectModel) -> float:
    return value

def generic_inverse(value: float, model: EffectModel) -> float:
    """
    Some sp_effect fields are hardcoded to be subtractable/divisible. Because of this,
    their values are negative in the params. Specifying this parser will reverse
    these values on per effect basis to have them make more sense.

    Examples:
    - Blessed Dew Talisman's `changeHpPoint` is set to -2 when it should heal
    - Malenia - Scarlet Rot's `changeHpPoint` is set to 26 when it should damage
    """
    return -value if model == EffectModel.ADDITIVE else _floor_decimal_2(2 - value)

def generic_inverse_percentage(value: float, model: EffectModel) -> float:
    """
    Flat percentage value, but also inverse. Turn -10 into 1.1 (ex. Assassin's Crimson Dagger).
    """
    return 1 + (-value / 100.0)

def poise(value: float, model: EffectModel) -> float:
    """
    Poise value is given as a "poise damage absorption" value, eg. `0.75`.
    The actual poise increase that is shown in game is the inverse of that:
        1 / .75 -> +33% poise.
    """
    return math.floor(1 / value * 100) / 100.0

def item_discovery(value: float, model: EffectModel) -> float:
    return value * 100


from enum import StrEnum
from typing import Any, Self, NamedTuple

from erdb.table._retrievers import RetrieverData
from erdb.table._common import TableSpec
from erdb.table.ammo import AmmoTableSpec
from erdb.table.armaments import ArmamentTableSpec
from erdb.table.armor import ArmorTableSpec
from erdb.table.ashes_of_war import AshOfWarTableSpec
from erdb.table.bolstering_materials import BolsteringMaterialTableSpec
from erdb.table.correction_attack import CorrectionAttackTableSpec
from erdb.table.correction_graph import CorrectionGraphTableSpec
from erdb.table.crafting_materials import CraftingMaterialTableSpec
from erdb.table.gestures import GestureTableSpec
from erdb.table.info import InfoTableSpec
from erdb.table.keys import KeyTableSpec
from erdb.table.reinforcements import ReinforcementTableSpec
from erdb.table.shop import ShopTableSpec
from erdb.table.spells import SpellTableSpec
from erdb.table.spirit_ashes import SpiritAshTableSpec
from erdb.table.talismans import TalismanTableSpec
from erdb.table.tools import ToolTableSpec
from erdb.typing.game_version import GameVersion
from erdb.typing.api_version import ApiVersion
from erdb.typing.params import ParamRow


class Generator(NamedTuple):
    spec: TableSpec
    data: RetrieverData

    def generate(self, api: ApiVersion | None = None) -> dict:
        api = self.spec.latest_api() if api is None else api

        def key(row: ParamRow) -> Any:
            return self.spec.get_pk(self.data, row)

        def value(row: ParamRow) -> Any:
            return self.spec.make_object(api, self.data, row)

        def valid(row: ParamRow) -> bool:
            return all(pred(row) for pred in self.spec.predicates)

        rows = self.data.main_param.values()
        return {key(row): value(row) for row in rows if valid(row)}

    @classmethod
    def create(cls, spec: TableSpec, version: GameVersion) -> Self:
        def retrieve_dict(retrievers: dict):
            return {field: retrievers[field].get(version) for field in retrievers.keys()}

        return cls(
            spec,
            RetrieverData(
                spec.main_param_retriever.get(version),
                retrieve_dict(spec.param_retrievers),
                retrieve_dict(spec.msg_retrievers),
                retrieve_dict(spec.shop_retrievers),
                spec.contrib_retriever.get(spec.title(), version),
            )
        )

class Table(StrEnum):
    ALL = "all"
    AMMO = "ammo"
    ARMAMENTS = "armaments"
    ARMOR = "armor"
    ASHES_OF_WAR = "ashes-of-war"
    BOLSTERING_MATERIALS = "bolstering-materials"
    CORRECTION_ATTACK = "correction-attack"
    CORRECTION_GRAPH = "correction-graph"
    CRAFTING_MATERIALS = "crafting-materials"
    GESTURES = "gestures"
    INFO = "info"
    KEYS = "keys"
    REINFORCEMENTS = "reinforcements"
    SHOP = "shop"
    SPELLS = "spells"
    SPIRIT_ASHES = "spirit-ashes"
    TASLISMANS = "talismans"
    TOOLS = "tools"

    def __str__(self):
        return self.value

    def __lt__(self, other: Self):
        assert isinstance(other, Table)
        return self.value < other.value

    def make_generator(self, version: GameVersion) -> Generator:
        return Generator.create(self.spec, version)

    @property
    def spec(self) -> TableSpec:
        return {
            Table.AMMO: AmmoTableSpec,
            Table.ARMAMENTS: ArmamentTableSpec,
            Table.ARMOR: ArmorTableSpec,
            Table.ASHES_OF_WAR: AshOfWarTableSpec,
            Table.BOLSTERING_MATERIALS: BolsteringMaterialTableSpec,
            Table.CORRECTION_ATTACK: CorrectionAttackTableSpec,
            Table.CORRECTION_GRAPH: CorrectionGraphTableSpec,
            Table.CRAFTING_MATERIALS: CraftingMaterialTableSpec,
            Table.GESTURES: GestureTableSpec,
            Table.INFO: InfoTableSpec,
            Table.KEYS: KeyTableSpec,
            Table.REINFORCEMENTS: ReinforcementTableSpec,
            Table.SHOP: ShopTableSpec,
            Table.SPELLS: SpellTableSpec,
            Table.SPIRIT_ASHES: SpiritAshTableSpec,
            Table.TASLISMANS: TalismanTableSpec,
            Table.TOOLS: ToolTableSpec,
        }[self]

    @property
    def param_name(self) -> str:
        return self.spec.main_param_retriever.param_name

    @property
    def title(self) -> str:
        return str(self).replace("-", " ").title()

    @classmethod
    def effective(cls) -> list[Self]:
        s = set(Table)
        s.remove(Table.ALL)
        return list(s)


from pydantic.dataclasses import dataclass

from erdb.typing.models import dt_config


@dataclass(config=dt_config())
class AttributesCorrection:
    strength: bool
    dexterity: bool
    intelligence: bool
    faith: bool
    arcane: bool

@dataclass(config=dt_config())
class AttributesOverride:
    strength: float | None = None
    dexterity: float | None = None
    intelligence: float | None = None
    faith: float | None = None
    arcane: float | None = None

@dataclass(config=dt_config())
class AttributesRatio:
    strength: float
    dexterity: float
    intelligence: float
    faith: float
    arcane: float

@dataclass(config=dt_config())
class Correction:
    physical: AttributesCorrection
    magic: AttributesCorrection
    fire: AttributesCorrection
    lightning: AttributesCorrection
    holy: AttributesCorrection

    @staticmethod
    def get_field_type():
        return AttributesCorrection

    @staticmethod
    def get_property(attribute: str, damage_type: str) -> str:
        return f"is{attribute}Correct_by{damage_type}"

@dataclass(config=dt_config())
class Override:
    physical: AttributesOverride | None = None
    magic: AttributesOverride | None = None
    fire: AttributesOverride | None = None
    lightning: AttributesOverride | None = None
    holy: AttributesOverride | None = None

    @staticmethod
    def get_field_type():
        return AttributesOverride

    @staticmethod
    def get_property(attribute: str, damage_type: str) -> str:
        return f"overwrite{attribute}CorrectRate_by{damage_type}"

@dataclass(config=dt_config())
class Ratio:
    physical: AttributesRatio
    magic: AttributesRatio
    fire: AttributesRatio
    lightning: AttributesRatio
    holy: AttributesRatio

    @staticmethod
    def get_property(attribute: str, damage_type: str) -> str:
        return f"Influence{attribute}CorrectRate_by{damage_type}"

    @staticmethod
    def get_field_type():
        return AttributesRatio

@dataclass(config=dt_config())
class CorrectionAttack:
    correction: Correction
    override: Override
    ratio: Ratio


details = {
    "iconId": {
        "DisplayName": "Icon ID",
        "Description": "Icon ID (When -1, no icon is required)",
        "Minimum": "-1",
        "Maximum": "999999",
        "SortID": "1000"
    },
    "conditionHp": {
        "DisplayName": "Trigger at HP Below %",
        "Description": "Set what percentage of maxHP the remaining HP will be activated",
        "Minimum": "-1",
        "Maximum": "100",
        "Increment": "0.1",
        "SortID": "18000"
    },
    "effectEndurance": {
        "DisplayName": "Duration",
        "Description": "Change duration / -1 for permanent / 0 for one moment only",
        "Minimum": "-1",
        "Maximum": "9999",
        "Increment": "0.1",
        "SortID": "19000"
    },
    "motionInterval": {
        "DisplayName": "Trigger Interval",
        "Description": "Set how many seconds it occurs",
        "Minimum": "-1",
        "Maximum": "999",
        "Increment": "0.1",
        "SortID": "20000"
    },
    "maxHpRate": {
        "DisplayName": "Max HP %",
        "Description": "Correct the maximum HP",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "21000"
    },
    "maxMpRate": {
        "DisplayName": "Max FP %",
        "Description": "Correct the maximum MP",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "22000"
    },
    "maxStaminaRate": {
        "DisplayName": "Max Stamina %",
        "Description": "Correct the maximum SP",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "23000"
    },
    "slashDamageCutRate": {
        "DisplayName": "Absorption %: Slash",
        "Description": "Slash damage ratio: Correct the calculated damage by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "29100"
    },
    "blowDamageCutRate": {
        "DisplayName": "Absorption %: Strike",
        "Description": "Batter damage ratio: The calculated damage is corrected by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "29200"
    },
    "thrustDamageCutRate": {
        "DisplayName": "Absorption %: Thrust",
        "Description": "Puncture damage ratio: Correct the calculated damage by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "29300"
    },
    "neutralDamageCutRate": {
        "DisplayName": "Absorption %: Standard",
        "Description": "Non-attribute damage ratio: Correct the calculated damage by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "29400"
    },
    "magicDamageCutRate": {
        "DisplayName": "Absorption %: Magic",
        "Description": "Magic damage multiplier: The calculated damage is corrected by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "30000"
    },
    "fireDamageCutRate": {
        "DisplayName": "Absorption %: Fire",
        "Description": "Flame damage ratio: Correct the calculated damage by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "31000"
    },
    "thunderDamageCutRate": {
        "DisplayName": "Absorption %: Lightning",
        "Description": "Electric shock damage ratio: Correct the calculated damage by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "31100"
    },
    "physicsAttackRate": {
        "DisplayName": "Attack %: Physical",
        "Description": "Physical damage ratio: Correct the calculated damage by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "32000"
    },
    "magicAttackRate": {
        "DisplayName": "Attack %: Magic",
        "Description": "Magic damage multiplier: The calculated damage is corrected by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "33000"
    },
    "fireAttackRate": {
        "DisplayName": "Attack %: Fire",
        "Description": "Flame damage ratio: Correct the calculated damage by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "34000"
    },
    "thunderAttackRate": {
        "DisplayName": "Attack %: Lightning",
        "Description": "Electric shock damage ratio: Correct the calculated damage by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "34100"
    },
    "physicsAttackPowerRate": {
        "DisplayName": "Power %: Physical",
        "Description": "Multiply the physical attack power by the set value",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "35000"
    },
    "magicAttackPowerRate": {
        "DisplayName": "Power %: Magic",
        "Description": "Multiply the magic attack power by the set value",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "36000"
    },
    "fireAttackPowerRate": {
        "DisplayName": "Power %: Fire",
        "Description": "Multiply the fire attack power by the set value",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "37000"
    },
    "thunderAttackPowerRate": {
        "DisplayName": "Power %: Lightning",
        "Description": "Multiply the electric shock attack power by the set value",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "37100"
    },
    "physicsAttackPower": {
        "DisplayName": "Damage +: Physical",
        "Description": "Add or subtract the value set for physical attack power",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "38000"
    },
    "magicAttackPower": {
        "DisplayName": "Damage +: Magic",
        "Description": "Add or subtract the value set for the magic attack power",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "39000"
    },
    "fireAttackPower": {
        "DisplayName": "Damage +: Fire",
        "Description": "Add or subtract the value set for the flame attack power",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "40000"
    },
    "thunderAttackPower": {
        "DisplayName": "Damage +: Lighting",
        "Description": "Add or subtract the value set for the electric shock attack power",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "40100"
    },
    "physicsDiffenceRate": {
        "DisplayName": "Defence %: Physical",
        "Description": "Multiply the set value for physical defense",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "41000"
    },
    "magicDiffenceRate": {
        "DisplayName": "Defence %: Magic",
        "Description": "Multiply the value set for magic defense",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "42000"
    },
    "fireDiffenceRate": {
        "DisplayName": "Defence %: Fire",
        "Description": "Multiply the fire defense by the set value",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "43000"
    },
    "thunderDiffenceRate": {
        "DisplayName": "Defence %: Lightning",
        "Description": "Multiply the value set for the electric shock defense",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "43100"
    },
    "physicsDiffence": {
        "DisplayName": "Defence +: Physical",
        "Description": "Add or subtract the value set for physical defense",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "44000"
    },
    "magicDiffence": {
        "DisplayName": "Defence +: Magic",
        "Description": "Add or subtract the value set for magic defense",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "45000"
    },
    "fireDiffence": {
        "DisplayName": "Defence +: Fire",
        "Description": "Add or subtract the value set for the fire defense",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "46000"
    },
    "thunderDiffence": {
        "DisplayName": "Defence +: Lightning",
        "Description": "Add or subtract the value set for the electric shock defense",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "46100"
    },
    "NoGuardDamageRate": {
        "DisplayName": "No Guard Damage %",
        "Description": "Replace the damage ratio at the time of the gap with the set value (set on the damage side)",
        "Minimum": "-100",
        "Maximum": "100",
        "SortID": "47000"
    },
    "vitalSpotChangeRate": {
        "DisplayName": "Vital Spot Change %",
        "Description": "Replaces the damage calculation of the sweet spot with the specified value (key point damage correction) -Invalid at 1.0",
        "Minimum": "-1",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "48000"
    },
    "normalSpotChangeRate": {
        "DisplayName": "Normal Hit Change %",
        "Description": "Replaces normal hit damage calculation with the specified number -Invalid at 1.0",
        "Minimum": "-1",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "49000"
    },
    "lookAtTargetPosOffset": {
        "DisplayName": "Look-At Target Position Offset",
        "Description": "Offset the target position when the enemy looks at. Set to crouch or mount on the side to be seen",
        "Minimum": "-99.99",
        "Maximum": "99.99",
        "SortID": "75600"
    },
    "behaviorId": {
        "DisplayName": "Behavior ID",
        "Description": "Specified -1 when dealing damage using action ID from special effects\nNPC Behavior - Requires State Info 142\nPlayer Behavior - Requires State Info 275\nBullet Behavior - Requires State Info 33",
        "Minimum": "-1",
        "Maximum": "2100000000",
        "SortID": "50100",
        "ParamRef1": "BehaviorParam_PC",
        "ParamRef2": "BehaviorParam"
    },
    "changeHpRate": {
        "DisplayName": "Current HP %",
        "Description": "Set what percentage of the maximum HP to subtract (or add) with one activation",
        "Minimum": "-100",
        "Maximum": "100",
        "Increment": "0.1",
        "SortID": "51000"
    },
    "changeHpPoint": {
        "DisplayName": "Current HP +",
        "Description": "Set how many points to subtract (or add) with one activation",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "52000"
    },
    "changeMpRate": {
        "DisplayName": "Current FP %",
        "Description": "Set what percentage of the maximum MP to subtract (or add) with one activation",
        "Minimum": "-100",
        "Maximum": "100",
        "Increment": "0.1",
        "SortID": "53000"
    },
    "changeMpPoint": {
        "DisplayName": "Current FP +",
        "Description": "Set how many points to subtract (or add) with one activation",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "54000"
    },
    "mpRecoverChangeSpeed": {
        "DisplayName": "FP Recovery Speed",
        "Description": "Change the recovery speed. Add or subtract to the standard recovery speed and initial recovery speed of the recovery formula.",
        "Minimum": "-100",
        "Maximum": "100",
        "SortID": "55000"
    },
    "changeStaminaRate": {
        "DisplayName": "Current Stamina %",
        "Description": "Set what percentage of the maximum stamina to subtract (or add) with one activation",
        "Minimum": "-100",
        "Maximum": "100",
        "Increment": "1",
        "SortID": "56000"
    },
    "changeStaminaPoint": {
        "DisplayName": "Current Stamina +",
        "Description": "Set how many points to subtract (or add) with one activation",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "57000"
    },
    "staminaRecoverChangeSpeed": {
        "DisplayName": "Stamina Recovery Speed",
        "Description": "Change the recovery speed. Add or subtract to the standard recovery speed and initial recovery speed of the recovery formula.",
        "Minimum": "-100",
        "Maximum": "100",
        "SortID": "58000"
    },
    "magicEffectTimeChange": {
        "DisplayName": "Magic Effect Time Change",
        "Description": "Add / subtract the time set for the effect duration only for magic that has the effect duration set to 0.1 seconds or more.",
        "Minimum": "-999",
        "Maximum": "999",
        "Increment": "0.1",
        "SortID": "59000"
    },
    "insideDurability": {
        "DisplayName": "Current Durability +",
        "Description": "Add or subtract the numerical value to the internal wear degree",
        "Minimum": "-999",
        "Maximum": "999",
        "SortID": "60000"
    },
    "maxDurability": {
        "DisplayName": "Max Durability",
        "Description": "Add the set value to the maximum value of the internal wear degree of durability.",
        "Minimum": "0",
        "Maximum": "999",
        "SortID": "61000"
    },
    "staminaAttackRate": {
        "DisplayName": "Attack %: Stamina",
        "Description": "Multiply the stamina attack power by a factor (1.0 1 times 0.5 half)",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "62000"
    },
    "poizonAttackPower": {
        "DisplayName": "Aux Inflict +: Poison",
        "Description": "A value to be added to the target's [poison resistance value] when it hits\nRequires State Info 2",
        "Minimum": "-99999",
        "Maximum": "99999",
        "SortID": "64000"
    },
    "diseaseAttackPower": {
        "DisplayName": "Aux Inflict +: Scarlet Rot",
        "Description": "Numerical value to be added to the target [Plague resistance value] when hit\nRequires State Info 5",
        "Minimum": "-99999",
        "Maximum": "99999",
        "SortID": "65000"
    },
    "bloodAttackPower": {
        "DisplayName": "Aux Inflict +: Hemorrhage",
        "Description": "A value to be added to the target's [bleeding resistance value] when it hits\nRequires State Info 6",
        "Minimum": "-99999",
        "Maximum": "99999",
        "SortID": "66000"
    },
    "curseAttackPower": {
        "DisplayName": "Aux Inflict +: Blight",
        "Description": "A number to be added to the target [curse resistance value] when hit\nRequires State Info 116",
        "Minimum": "-99999",
        "Maximum": "99999",
        "SortID": "66100"
    },
    "fallDamageRate": {
        "DisplayName": "Fall Damage %",
        "Description": "Multiply the damage calculation when falling",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "67000"
    },
    "soulRate": {
        "DisplayName": "Rune Gain %",
        "Description": "The amount of soul acquired when defeating an enemy is added by the specified multiple.\nRequires State Info 76",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "68000"
    },
    "equipWeightChangeRate": {
        "DisplayName": "Max Equip Load %",
        "Description": "Multiply the maximum equipment weight by the set magnification",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "69000"
    },
    "allItemWeightChangeRate": {
        "DisplayName": "Item Weight %",
        "Description": "Multiply the maximum weight you have by the set magnification",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "70000"
    },
    "soul": {
        "DisplayName": "Rune Gain +",
        "Description": "Add the set value to the possessed soul",
        "Minimum": "-99999999",
        "Maximum": "2100000000",
        "SortID": "71000"
    },
    "animIdOffset": {
        "DisplayName": "Anim ID Offset",
        "Description": "Anime ID offset (invalid-1)",
        "Minimum": "-1",
        "Maximum": "2100000000",
        "SortID": "111500"
    },
    "haveSoulRate": {
        "DisplayName": "Rune Possession Amount %",
        "Description": "For enemy lap effect. It is applied when the soul goes out from the set character.",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "72000"
    },
    "targetPriority": {
        "DisplayName": "Target Priority",
        "Description": "During multiplayer, the enemy will give priority to being targeted as a target. Addition of priority. 0 is the default. It will be often targeted with a positive value. Minus is up to -1.",
        "Minimum": "-1",
        "Maximum": "10",
        "SortID": "73000"
    },
    "sightSearchEnemyRate": {
        "DisplayName": "Enemy Vision Adjustment",
        "Description": "Correct the ease of finding by a magnification",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "73620"
    },
    "hearingSearchEnemyRate": {
        "DisplayName": "Enemy Listen Adjustment",
        "Description": "Correct the loudness of the AI sound emitted by the magnification",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "73920"
    },
    "grabityRate": {
        "DisplayName": "Gravity %",
        "Description": "Gravity rate",
        "Minimum": "0",
        "Maximum": "1",
        "Increment": "0.001",
        "SortID": "87000"
    },
    "registPoizonChangeRate": {
        "DisplayName": "Aux Resist %: Poison",
        "Description": "Multiply the poison resistance value by the set multiplier",
        "Minimum": "-1",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "88000"
    },
    "registDiseaseChangeRate": {
        "DisplayName": "Aux Resist %: Scarlet Rot",
        "Description": "Multiply the plague resistance value by the set magnification",
        "Minimum": "-1",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "88100"
    },
    "registBloodChangeRate": {
        "DisplayName": "Aux Resist %: Hemorrhage",
        "Description": "Multiply the bleeding resistance value by the set magnification",
        "Minimum": "-1",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "88200"
    },
    "registCurseChangeRate": {
        "DisplayName": "Aux Resist %: Blight",
        "Description": "Multiply the spell resistance value by the set multiplier",
        "Minimum": "-1",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "88300"
    },
    "soulStealRate": {
        "DisplayName": "Soul Steal %",
        "Description": "Defense against HP robbed by NPCs in Soul Steel",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "91000"
    },
    "lifeReductionRate": {
        "DisplayName": "Effect Duration %",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "100000"
    },
    "hpRecoverRate": {
        "DisplayName": "HP Recovery Effectiveness %",
        "Description": "It doesn't work when HP decreases.",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "101000"
    },
    "replaceSpEffectId": {
        "DisplayName": "Chain SpEffect ID",
        "Description": "Special effect ID added at the end of life (-1 is ignored)",
        "Minimum": "-1",
        "Maximum": "2100000000",
        "SortID": "111000",
        "ParamRef1": "SpEffectParam"
    },
    "cycleOccurrenceSpEffectId": {
        "DisplayName": "Cycle SpEffect ID",
        "Description": "Special effect ID that occurs in each activation cycle (-1 is ignored)",
        "Minimum": "-1",
        "Maximum": "2100000000",
        "SortID": "111010",
        "ParamRef1": "SpEffectParam"
    },
    "atkOccurrenceSpEffectId": {
        "DisplayName": "Attack SpEffect ID",
        "Description": "Special effect ID that occurs when hitting an attack (-1 is ignored)",
        "Minimum": "-1",
        "Maximum": "2100000000",
        "SortID": "111020",
        "ParamRef1": "SpEffectParam"
    },
    "guardDefFlickPowerRate": {
        "DisplayName": "Guard Defense - Weapon Repel Power %",
        "Description": "Repellent defense correction value when guarding",
        "DisplayFormat": "%.2f",
        "Minimum": "-99",
        "Maximum": "99",
        "SortID": "27530"
    },
    "guardStaminaCutRate": {
        "DisplayName": "Guard Stability %",
        "Description": "Stamina cut rate correction value when guarding",
        "DisplayFormat": "%.2f",
        "Minimum": "-99",
        "Maximum": "99",
        "SortID": "62100"
    },
    "rayCastPassedTime": {
        "DisplayName": "Raycast Passed Time",
        "Description": "Passing the line of sight: Activation time [ms] (for evil eye)",
        "Minimum": "-1",
        "Maximum": "30000",
        "SortID": "108000"
    },
    "magicSubCategoryChange1": {
        "DisplayName": "Conditional Category [1]",
        "Enum": "ATK_SUB_CATEGORY",
        "Description": "Vs to subcategory parameter change 1",
        "SortID": "26300"
    },
    "magicSubCategoryChange2": {
        "DisplayName": "Conditional Category [2]",
        "Enum": "ATK_SUB_CATEGORY",
        "Description": "Vs to subcategory parameter change 2",
        "SortID": "26310"
    },
    "bowDistRate": {
        "DisplayName": "Bow Distance %",
        "Description": "Correction value added to the flight distance correction of the weapon",
        "Minimum": "-100",
        "Maximum": "999",
        "SortID": "62200"
    },
    "spCategory": {
        "DisplayName": "Effect Category",
        "Enum": "SP_EFFECT_SPCATEGORY",
        "Description": "Categories that determine behavior such as overwriting special effects",
        "Maximum": "60000",
        "SortID": "2100"
    },
    "categoryPriority": {
        "DisplayName": "Category Priority",
        "Description": "Priority within the same category (lower one has priority)",
        "SortID": "2200"
    },
    "saveCategory": {
        "DisplayName": "Save Category",
        "Enum": "SP_EFFECT_SAVE_CATEGORY",
        "Description": "Category to save special effects",
        "Minimum": "-1",
        "SortID": "2300"
    },
    "changeMagicSlot": {
        "DisplayName": "Magic Attunement Slot",
        "Description": "You can increase the specified number of magic registration slots",
        "Maximum": "3",
        "SortID": "83000"
    },
    "changeMiracleSlot": {
        "DisplayName": "Miracle Attunement Slot",
        "Description": "You can increase the specified number of trajectory registration frames.",
        "Maximum": "3",
        "SortID": "84000"
    },
    "heroPointDamage": {
        "DisplayName": "Humanity Damage",
        "Description": "Damage value given to human nature value",
        "Minimum": "-99",
        "Maximum": "99",
        "SortID": "112000"
    },
    "defFlickPower": {
        "DisplayName": "Weapon Repel Defence +",
        "Description": "Set a value that overwrites the repelling defense",
        "SortID": "27510"
    },
    "flickDamageCutRate": {
        "DisplayName": "Weapon Repel Absorption %",
        "Description": "Set a value that overwrites the damage attenuation rate at the time of repelling",
        "Maximum": "100",
        "SortID": "27520"
    },
    "bloodDamageRate": {
        "DisplayName": "Damage Correction %: Hemorrhage",
        "Description": "Point damage of state change type [bleeding], correction value used only when% damage",
        "SortID": "90000"
    },
    "dmgLv_None": {
        "DisplayName": "Damage Level: No Impact",
        "Enum": "ATKPARAM_REP_DMGTYPE",
        "Description": "Specify the type to replace the damage Lv0",
        "Minimum": "0",
        "SortID": "27110"
    },
    "dmgLv_S": {
        "DisplayName": "Damage Level: Small Impact",
        "Enum": "ATKPARAM_REP_DMGTYPE",
        "Description": "Specify the type to replace the damage Lv1",
        "Minimum": "0",
        "SortID": "27130"
    },
    "dmgLv_M": {
        "DisplayName": "Damage Level: Medium Impact",
        "Enum": "ATKPARAM_REP_DMGTYPE",
        "Description": "Specify the type to replace the damage Lv2",
        "Minimum": "0",
        "SortID": "27140"
    },
    "dmgLv_L": {
        "DisplayName": "Damage Level: Large Impact",
        "Enum": "ATKPARAM_REP_DMGTYPE",
        "Description": "Specify the type to replace the damage Lv3",
        "Minimum": "0",
        "SortID": "27150"
    },
    "dmgLv_BlowM": {
        "DisplayName": "Damage Level: Medium Blow",
        "Enum": "ATKPARAM_REP_DMGTYPE",
        "Description": "Specify the type to replace the damage Lv4",
        "Minimum": "0",
        "SortID": "27170"
    },
    "dmgLv_Push": {
        "DisplayName": "Damage Level: Push",
        "Enum": "ATKPARAM_REP_DMGTYPE",
        "Description": "Specify the type to replace the damage Lv5",
        "Minimum": "0",
        "SortID": "27200"
    },
    "dmgLv_Strike": {
        "DisplayName": "Damage Level: Strike",
        "Enum": "ATKPARAM_REP_DMGTYPE",
        "Description": "Specify the type to replace the damage Lv6",
        "Minimum": "0",
        "SortID": "27180"
    },
    "dmgLv_BlowS": {
        "DisplayName": "Damage Level: Small Blow",
        "Enum": "ATKPARAM_REP_DMGTYPE",
        "Description": "Specify the type to replace the damage Lv7",
        "Minimum": "0",
        "SortID": "27160"
    },
    "dmgLv_Min": {
        "DisplayName": "Damage Level: Minimal",
        "Enum": "ATKPARAM_REP_DMGTYPE",
        "Description": "Specify the type to replace the damage Lv8",
        "Minimum": "0",
        "SortID": "27120"
    },
    "dmgLv_Uppercut": {
        "DisplayName": "Damage Level: Uppercut",
        "Enum": "ATKPARAM_REP_DMGTYPE",
        "Description": "Specify the type to replace the damage Lv9",
        "Minimum": "0",
        "SortID": "27190"
    },
    "dmgLv_BlowLL": {
        "DisplayName": "Damage Level: Blow Backward",
        "Enum": "ATKPARAM_REP_DMGTYPE",
        "Description": "Specify the type to replace the damage Lv10",
        "Minimum": "0",
        "SortID": "27200"
    },
    "dmgLv_Breath": {
        "DisplayName": "Damage Level: Breath",
        "Enum": "ATKPARAM_REP_DMGTYPE",
        "Description": "Specify the type to replace the damage Lv11",
        "Minimum": "0",
        "SortID": "27210"
    },
    "atkAttribute": {
        "DisplayName": "Attack Attribute",
        "Enum": "ATKPARAM_ATKATTR_TYPE",
        "Description": "Physical attributes to set for special effects",
        "SortID": "27400"
    },
    "spAttribute": {
        "DisplayName": "SpEffect Attribute",
        "Enum": "ATKPARAM_SPATTR_TYPE",
        "Description": "Special attributes to set for special effects",
        "SortID": "27500"
    },
    "stateInfo": {
        "DisplayName": "State Info",
        "Enum": "SP_EFFECT_TYPE",
        "Description": "State change judgment flag",
        "Maximum": "60000",
        "SortID": "2000"
    },
    "wepParamChange": {
        "DisplayName": "Weapon Parameter Change",
        "Enum": "SP_EFE_WEP_CHANGE_PARAM",
        "Description": "Specify which weapon is effective. If there is no limit, all attacks and defenses including enemies are targeted",
        "SortID": "24000"
    },
    "moveType": {
        "DisplayName": "Move Type",
        "Enum": "SP_EFFECT_MOVE_TYPE",
        "Description": "Movement type. Change the movement speed.",
        "SortID": "85000"
    },
    "lifeReductionType": {
        "DisplayName": "Effect Duration Multiplier - State Info",
        "Enum": "SP_EFFECT_TYPE",
        "Maximum": "255",
        "SortID": "99000"
    },
    "throwCondition": {
        "DisplayName": "Throw Trigger",
        "Enum": "SP_EFFECT_THROW_CONDITION_TYPE",
        "Description": "Throwing conditions. Affects the throwing mask.",
        "SortID": "110000"
    },
    "addBehaviorJudgeId_condition": {
        "DisplayName": "Add Behavior Judge ID - Trigger",
        "Description": "Condition value to add a value to the action judgment ID (Def: -1)",
        "Minimum": "-1",
        "Maximum": "9",
        "SortID": "113000"
    },
    "freezeDamageRate": {
        "DisplayName": "Damage Correction %: Frostbite",
        "Description": "Correction value used only for point damage and% damage of state change type [cold air]",
        "SortID": "90100"
    },
    "effectTargetSelf": {
        "DisplayName": "Target Self",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Only the target for which this judgment is checked is effective, the default is ×",
        "Maximum": "1",
        "SortID": "3000"
    },
    "effectTargetFriend": {
        "DisplayName": "Target Friend",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Only the target for which this judgment is checked is effective, the default is ×",
        "Maximum": "1",
        "SortID": "4000"
    },
    "effectTargetEnemy": {
        "DisplayName": "Target Enemy",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Only the target for which this judgment is checked is effective, the default is ×",
        "Maximum": "1",
        "SortID": "5000"
    },
    "effectTargetPlayer": {
        "DisplayName": "Target Player",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Only the target for which this judgment is checked is effective, the default is ×",
        "Maximum": "1",
        "SortID": "6000"
    },
    "effectTargetAI": {
        "DisplayName": "Target AI",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Only the target for which this judgment is checked is effective, the default is ×",
        "Maximum": "1",
        "SortID": "7000"
    },
    "effectTargetLive": {
        "DisplayName": "Target Live",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Only the target for which this judgment is checked is effective, the default is ×",
        "Maximum": "1",
        "SortID": "8000"
    },
    "effectTargetGhost": {
        "DisplayName": "Target Ghost",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Only the target for which this judgment is checked is effective, the default is ×",
        "Maximum": "1",
        "SortID": "9000"
    },
    "disableSleep": {
        "DisplayName": "Disable Sleep",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "If this effect is applied, you will not sleep",
        "Maximum": "1",
        "SortID": "104070"
    },
    "disableMadness": {
        "DisplayName": "Disable Madness",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "With this effect, you won't go mad",
        "Maximum": "1",
        "SortID": "104080"
    },
    "effectTargetAttacker": {
        "DisplayName": "Target Attacker",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Apply special effects to attackers after damage (cannot enter defenders)",
        "Maximum": "1",
        "SortID": "12000"
    },
    "dispIconNonactive": {
        "DisplayName": "Display Icon when Inactive",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "The icon is displayed even in the state of waiting for activation.",
        "Maximum": "1",
        "SortID": "13000"
    },
    "regainGaugeDamage": {
        "DisplayName": "Generate Regain Gauge",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Whether to generate a regain gauge",
        "Maximum": "1",
        "SortID": "161000"
    },
    "bAdjustMagicAblity": {
        "DisplayName": "Use Intelligence Correction",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Do you want to correct the magic power?",
        "Maximum": "1",
        "SortID": "15000"
    },
    "bAdjustFaithAblity": {
        "DisplayName": "Use Faith Correction",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Do you correct your faith?",
        "Maximum": "1",
        "SortID": "16000"
    },
    "bGameClearBonus": {
        "DisplayName": "Enable for Game Clear",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Whether it is for the game clear lap bonus.",
        "Maximum": "1",
        "SortID": "17000"
    },
    "magParamChange": {
        "DisplayName": "Affects Sorcery",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Set whether or not it is effective against magic",
        "Maximum": "1",
        "SortID": "25000"
    },
    "miracleParamChange": {
        "DisplayName": "Affects Incantation",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Set whether or not it is effective against miracles",
        "Maximum": "1",
        "SortID": "26000"
    },
    "clearSoul": {
        "DisplayName": "Clear Held Runes",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Possession soul becomes 0.",
        "Maximum": "1",
        "SortID": "27000"
    },
    "requestSOS": {
        "DisplayName": "Request: Friendly Phantom Summon",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "If checked, issue an SOS sign request when activated",
        "Maximum": "1",
        "SortID": "76000"
    },
    "requestBlackSOS": {
        "DisplayName": "Request: Hostile Phantom Summon",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "If checked, issue a black SOS sign request when activated",
        "Maximum": "1",
        "SortID": "77000"
    },
    "requestForceJoinBlackSOS": {
        "DisplayName": "Request: Invasion",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "If checked, issue an intrusion_A request when activated",
        "Maximum": "1",
        "SortID": "78000"
    },
    "requestKickSession": {
        "DisplayName": "Request: Kick",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "If checked, a kick request will be issued at the time of activation.",
        "Maximum": "1",
        "SortID": "79000"
    },
    "requestLeaveSession": {
        "DisplayName": "Request: Return to Own World",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "If checked, an exit request will be issued at the time of activation.",
        "Maximum": "1",
        "SortID": "80000"
    },
    "requestNpcInveda": {
        "DisplayName": "Request: NPC Invasion",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "If checked, an intrusion request to the NPC will be issued at the time of activation.",
        "Maximum": "1",
        "SortID": "80100"
    },
    "noDead": {
        "DisplayName": "Is No Dead",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Whether you can be corpse. With this check, you will not be dead",
        "Maximum": "1",
        "SortID": "81000"
    },
    "bCurrHPIndependeMaxHP": {
        "DisplayName": "Current HP independent Max HP",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Does HP now affect even if the maximum HP increases or decreases?",
        "Maximum": "1",
        "SortID": "82000"
    },
    "corrosionIgnore": {
        "DisplayName": "Ignore Durability Loss",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "[State change type] ignores [Durability] decrease due to [Corrosion]",
        "Maximum": "1",
        "SortID": "92000"
    },
    "sightSearchCutIgnore": {
        "DisplayName": "Ignore Vision Reductions",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Ignore visual search invalidity",
        "Maximum": "1",
        "SortID": "93000"
    },
    "hearingSearchCutIgnore": {
        "DisplayName": "Ignore Listen Reductions",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Ignore auditory search invalidity",
        "Maximum": "1",
        "SortID": "94000"
    },
    "antiMagicIgnore": {
        "DisplayName": "Ignore Anti-Magic Effect",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "You can use magic even in the anti-magic range",
        "Maximum": "1",
        "SortID": "95000"
    },
    "fakeTargetIgnore": {
        "DisplayName": "Ignore Fake Targets",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Don't get caught in the fake target that occurred",
        "Maximum": "1",
        "SortID": "96000"
    },
    "fakeTargetIgnoreUndead": {
        "DisplayName": "Ignore Fake Human Targets",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "You will not be caught by the fake target of the human system that occurred",
        "Maximum": "1",
        "SortID": "96500"
    },
    "fakeTargetIgnoreAnimal": {
        "DisplayName": "Ignore Fake Animal Targets",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "You will not be caught by the fake target of the beast system that occurred",
        "Maximum": "1",
        "SortID": "96600"
    },
    "grabityIgnore": {
        "DisplayName": "Ignore Gravity Change",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Gravity effect disabled",
        "Maximum": "1",
        "SortID": "97000"
    },
    "disablePoison": {
        "DisplayName": "Disable Poison",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "If this effect is applied, it will not be poisoned.",
        "Maximum": "1",
        "SortID": "102000"
    },
    "disableDisease": {
        "DisplayName": "Disable Scarlet Rot",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "If this effect is applied, you will not get plague",
        "Maximum": "1",
        "SortID": "103000"
    },
    "disableBlood": {
        "DisplayName": "Disable Hemorrhage",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "With this effect, you won't get bleeding",
        "Maximum": "1",
        "SortID": "104000"
    },
    "disableCurse": {
        "DisplayName": "Disable Blight",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "With this effect, you won't be cursed",
        "Maximum": "1",
        "SortID": "104050"
    },
    "enableCharm": {
        "DisplayName": "Enable Charm",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "If this effect is applied, you will be fascinated.",
        "Maximum": "1",
        "SortID": "104100"
    },
    "enableLifeTime": {
        "DisplayName": "Enable Lifetime",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Will the life be extended when the flag is set by TAE?",
        "Maximum": "1",
        "SortID": "105000"
    },
    "bAdjustStrengthAblity": {
        "DisplayName": "Use Strength Correction",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Do you want to correct your strength?",
        "Maximum": "1",
        "SortID": "16100"
    },
    "bAdjustAgilityAblity": {
        "DisplayName": "Use Dexterity Correction",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Do you want to correct your workmanship?",
        "Maximum": "1",
        "SortID": "16200"
    },
    "eraseOnBonfireRecover": {
        "DisplayName": "Erase on Bonfire Rest",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Will it be extinguished by bonfire recovery?",
        "Maximum": "1",
        "SortID": "20600"
    },
    "throwAttackParamChange": {
        "DisplayName": "Throw Attack Param Change",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Set whether or not it is effective against throwing attacks",
        "Maximum": "1",
        "SortID": "26200"
    },
    "requestLeaveColiseumSession": {
        "DisplayName": "Request: Leave Arena",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "If checked, a request to leave the arena will be issued at the time of activation.",
        "Maximum": "1",
        "SortID": "80000"
    },
    "isExtendSpEffectLife": {
        "DisplayName": "Has Effect Duration Adjustment Applied",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Whether to be eligible for extension when the life extension effect is applied",
        "Maximum": "1",
        "SortID": "105100"
    },
    "hasTarget": {
        "DisplayName": "Has Target",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Do you know the enemy? : [Activation condition] (for evil eye users)",
        "Maximum": "1",
        "SortID": "109000"
    },
    "replanningOnFire": {
        "DisplayName": "Cancel on Fire Damage",
        "Enum": "BOOL_CIRCLECROSS_TYPE",
        "Description": "Whether to replan at the time of activation",
        "Maximum": "1",
        "SortID": "72700"
    },
    "vowType0": {
        "DisplayName": "Trigger for Vow Type [0]",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Pledge 0",
        "Maximum": "1",
        "SortID": "130000"
    },
    "vowType1": {
        "DisplayName": "Trigger for Vow Type [1]",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Pledge 1",
        "Maximum": "1",
        "SortID": "130100"
    },
    "vowType2": {
        "DisplayName": "Trigger for Vow Type [2]",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Pledge 2",
        "Maximum": "1",
        "SortID": "130200"
    },
    "vowType3": {
        "DisplayName": "Trigger for Vow Type [3]",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Pledge 3",
        "Maximum": "1",
        "SortID": "130300"
    },
    "vowType4": {
        "DisplayName": "Trigger for Vow Type [4]",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Pledge 4",
        "Maximum": "1",
        "SortID": "130400"
    },
    "vowType5": {
        "DisplayName": "Trigger for Vow Type [5]",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Pledge 5",
        "Maximum": "1",
        "SortID": "130500"
    },
    "vowType6": {
        "DisplayName": "Trigger for Vow Type [6]",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Pledge 6",
        "Maximum": "1",
        "SortID": "130600"
    },
    "vowType7": {
        "DisplayName": "Trigger for Vow Type [7]",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Pledge 7",
        "Maximum": "1",
        "SortID": "130700"
    },
    "vowType8": {
        "DisplayName": "Trigger for Vow Type [8]",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Pledge 8",
        "Maximum": "1",
        "SortID": "130800"
    },
    "vowType9": {
        "DisplayName": "Trigger for Vow Type [9]",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Pledge 9",
        "Maximum": "1",
        "SortID": "130900"
    },
    "vowType10": {
        "DisplayName": "Trigger for Vow Type [10]",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Pledge 10",
        "Maximum": "1",
        "SortID": "131000"
    },
    "vowType11": {
        "DisplayName": "Trigger for Vow Type [11]",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Pledge 11",
        "Maximum": "1",
        "SortID": "131100"
    },
    "vowType12": {
        "DisplayName": "Trigger for Vow Type [12]",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Pledge 12",
        "Maximum": "1",
        "SortID": "131200"
    },
    "vowType13": {
        "DisplayName": "Trigger for Vow Type [13]",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Pledge 13",
        "Maximum": "1",
        "SortID": "131300"
    },
    "vowType14": {
        "DisplayName": "Trigger for Vow Type [14]",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Pledge 14",
        "Maximum": "1",
        "SortID": "131400"
    },
    "vowType15": {
        "DisplayName": "Trigger for Vow Type [15]",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Pledge 15",
        "Maximum": "1",
        "SortID": "131500"
    },
    "repAtkDmgLv": {
        "DisplayName": "Change Attacker Damage Level",
        "Enum": "ATKPARAM_REP_DMGTYPE",
        "Description": "The damage level of the attacking side changes to this value",
        "Minimum": "0",
        "SortID": "152000"
    },
    "sightSearchRate": {
        "DisplayName": "AI Vision Correction",
        "Description": "Correct the ease of finding with a magnification\nRequires State Info 267",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "73720"
    },
    "effectTargetOpposeTarget": {
        "DisplayName": "Trigger for Opponent",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Only the target for which this judgment is checked is effective, the default is ×",
        "Maximum": "1",
        "SortID": "12100"
    },
    "effectTargetFriendlyTarget": {
        "DisplayName": "Trigger for Friendly",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Only the target for which this judgment is checked is effective, the default is ×",
        "Maximum": "1",
        "SortID": "12200"
    },
    "effectTargetSelfTarget": {
        "DisplayName": "Trigger for Self",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Only the target for which this judgment is checked is effective, the default is ×",
        "Maximum": "1",
        "SortID": "12300"
    },
    "effectTargetPcHorse": {
        "DisplayName": "Trigger for Horse",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Only the target for which this judgment is checked is effective, the default is ×",
        "Maximum": "1",
        "SortID": "12350"
    },
    "effectTargetPcDeceased": {
        "DisplayName": "Trigger for Dead Player",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Only the target for which this judgment is checked is effective, the default is ×",
        "Maximum": "1",
        "SortID": "12400"
    },
    "isContractSpEffectLife": {
        "DisplayName": "Can Duration be Reduced",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Whether it will be shortened when the life shortening effect is applied",
        "Maximum": "1",
        "SortID": "105110"
    },
    "isWaitModeDelete": {
        "DisplayName": "Delete in Wait Mode",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Do you want to delete it the moment you are in a waiting state?",
        "Maximum": "1",
        "SortID": "411000"
    },
    "isIgnoreNoDamage": {
        "DisplayName": "Apply through No Damage",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Whether to apply the damage from this special effect even in the invincible state only when the state change type \"Apply the activation function even when invincible\" is applied.",
        "Maximum": "1",
        "SortID": "411100"
    },
    "changeTeamType": {
        "DisplayName": "Team Type",
        "Enum": "SP_EFFECT_CHANGE_TEAM_TYPE",
        "Description": "Overwrites the specified team type\nRequires State Info 132",
        "Minimum": "-1",
        "Maximum": "99",
        "SortID": "84900"
    },
    "dmypolyId": {
        "DisplayName": "Bullet Behavior ID - DmyPoly ID",
        "Description": "Damipoli ID. Damipoli ID range is 0 to 999.1000, 10000 is the category number.",
        "Minimum": "-1",
        "Maximum": "31999",
        "SortID": "50200"
    },
    "vfxId": {
        "DisplayName": "Effect VFX [0]",
        "Description": "Special effect VfxId (-1 disabled)",
        "Minimum": "-1",
        "Maximum": "2100000000",
        "SortID": "14100",
        "ParamRef1": "SpEffectVfxParam"
    },
    "accumuOverFireId": {
        "DisplayName": "Accumulator - Over Value - SpEffect ID",
        "Description": "Special effect Id activated at the upper limit of the spirit ball",
        "Minimum": "-1",
        "Maximum": "2100000000",
        "SortID": "114200"
    },
    "accumuOverVal": {
        "DisplayName": "Accumulator - Over Value",
        "Description": "Genkitama upper limit",
        "Minimum": "-1",
        "Maximum": "2100000000",
        "SortID": "114100"
    },
    "accumuUnderFireId": {
        "DisplayName": "Accumulator - Under Value - SpEffect ID",
        "Description": "Special effect Id activated at the lower limit of the spirit ball",
        "Minimum": "-1",
        "Maximum": "2100000000",
        "SortID": "114400"
    },
    "accumuUnderVal": {
        "DisplayName": "Accumulator - Under Value",
        "Description": "Genkitama lower limit",
        "Minimum": "-1",
        "Maximum": "2100000000",
        "SortID": "114300"
    },
    "accumuVal": {
        "DisplayName": "Accumulator - Increment Value",
        "Description": "Genkitama accumulation value",
        "Minimum": "-99999999",
        "Maximum": "2100000000",
        "SortID": "114500"
    },
    "eye_angX": {
        "DisplayName": "Vision - Overwrite Height Angle",
        "Description": "Override the findability angle x",
        "Maximum": "180",
        "SortID": "73730"
    },
    "eye_angY": {
        "DisplayName": "Vision - Overwrite Width Angle",
        "Description": "Override the findability angle y",
        "Maximum": "180",
        "SortID": "73740"
    },
    "addDeceasedLv": {
        "DisplayName": "Add Deceased Level",
        "Description": "Add this value to the degree of death",
        "Minimum": "-999",
        "Maximum": "999",
        "SortID": "115000"
    },
    "vfxId1": {
        "DisplayName": "Effect VFX [1]",
        "Description": "Special effect VfxId1 (-1 invalid)",
        "Minimum": "-1",
        "Maximum": "2100000000",
        "SortID": "14110",
        "ParamRef1": "SpEffectVfxParam"
    },
    "vfxId2": {
        "DisplayName": "Effect VFX [2]",
        "Description": "Special effect VfxId2 (-1 invalid)",
        "Minimum": "-1",
        "Maximum": "2100000000",
        "SortID": "14120",
        "ParamRef1": "SpEffectVfxParam"
    },
    "vfxId3": {
        "DisplayName": "Effect VFX [3]",
        "Description": "Special effect VfxId3 (-1 invalid)",
        "Minimum": "-1",
        "Maximum": "2100000000",
        "SortID": "14130",
        "ParamRef1": "SpEffectVfxParam"
    },
    "vfxId4": {
        "DisplayName": "Effect VFX [4]",
        "Description": "Special effect VfxId4 (-1 invalid)",
        "Minimum": "-1",
        "Maximum": "2100000000",
        "SortID": "14140",
        "ParamRef1": "SpEffectVfxParam"
    },
    "vfxId5": {
        "DisplayName": "Effect VFX [5]",
        "Description": "Special effect VfxId5 (-1 invalid)",
        "Minimum": "-1",
        "Maximum": "2100000000",
        "SortID": "14150",
        "ParamRef1": "SpEffectVfxParam"
    },
    "vfxId6": {
        "DisplayName": "Effect VFX [6]",
        "Description": "Special effect VfxId6 (-1 invalid)",
        "Minimum": "-1",
        "Maximum": "2100000000",
        "SortID": "14160",
        "ParamRef1": "SpEffectVfxParam"
    },
    "vfxId7": {
        "DisplayName": "Effect VFX [7]",
        "Description": "Special effect VfxId7 (-1 invalid)",
        "Minimum": "-1",
        "Maximum": "2100000000",
        "SortID": "14170",
        "ParamRef1": "SpEffectVfxParam"
    },
    "freezeAttackPower": {
        "DisplayName": "Aux Inflict +: Frostbite",
        "Description": "Numerical value to be added to the target [Cold air resistance value] when hit\nRequires State Info 260",
        "Minimum": "-99999",
        "Maximum": "99999",
        "SortID": "66200"
    },
    "AppearAiSoundId": {
        "DisplayName": "Generated AI Sound ID",
        "Description": "Generates AI sound parameters with set values",
        "Minimum": "0",
        "Maximum": "2100000000",
        "SortID": "75200"
    },
    "addFootEffectSfxId": {
        "DisplayName": "Automatic Foot Effect - SFX ID Offset",
        "Description": "The identifier of the foot effect that is additionally generated during special effects. XYYZZZ ZZZ",
        "Minimum": "-1",
        "Maximum": "999",
        "SortID": "14500",
        "ParamRef1": "FootSfxParam"
    },
    "dexterityCancelSystemOnlyAddDexterity": {
        "DisplayName": "Cast Speed +",
        "Description": "Add this value when calculating the end timing of the TAE flag of \"Skill Cancel\".",
        "Minimum": "-99",
        "Maximum": "99",
        "SortID": "410000"
    },
    "teamOffenseEffectivity": {
        "DisplayName": "Team Attack Effectivity",
        "Description": "Overwrite and change the target [Team Attack Influence] value. Do not change the default value (-1).",
        "Minimum": "-1",
        "Maximum": "100",
        "SortID": "75300"
    },
    "toughnessDamageCutRate": {
        "DisplayName": "Toughness Absorption %",
        "Description": "Toughness version cut rate",
        "Minimum": "0",
        "Maximum": "10",
        "Increment": "0.001",
        "SortID": "90500"
    },
    "weakDmgRateA": {
        "DisplayName": "Damage Multiplier %: Type A",
        "Description": "Special Attack A Damage multiplier is corrected. 1 is normal.",
        "Minimum": "0",
        "Maximum": "99.999",
        "Increment": "0.001",
        "SortID": "46500"
    },
    "weakDmgRateB": {
        "DisplayName": "Damage Multiplier %: Type B",
        "Description": "Special attack B Damage multiplier is corrected. 1 is normal.",
        "Minimum": "0",
        "Maximum": "99.999",
        "Increment": "0.001",
        "SortID": "46510"
    },
    "weakDmgRateC": {
        "DisplayName": "Damage Multiplier %: Type C",
        "Description": "Special attack C Damage multiplier is corrected. 1 is normal.",
        "Minimum": "0",
        "Maximum": "99.999",
        "Increment": "0.001",
        "SortID": "46520"
    },
    "weakDmgRateD": {
        "DisplayName": "Damage Multiplier %: Type D",
        "Description": "Special attack D Damage multiplier is corrected. 1 is normal.",
        "Minimum": "0",
        "Maximum": "99.999",
        "Increment": "0.001",
        "SortID": "46530"
    },
    "weakDmgRateE": {
        "DisplayName": "Damage Multiplier %: Type E",
        "Description": "Special attack E Damage multiplier is corrected. 1 is normal.",
        "Minimum": "0",
        "Maximum": "99.999",
        "Increment": "0.001",
        "SortID": "46540"
    },
    "weakDmgRateF": {
        "DisplayName": "Damage Multiplier %: Type F",
        "Description": "Special attack F Damage multiplier is corrected. 1 is normal.",
        "Minimum": "0",
        "Maximum": "99.999",
        "Increment": "0.001",
        "SortID": "46550"
    },
    "darkDamageCutRate": {
        "DisplayName": "Absorption %: Holy",
        "Description": "Dark damage multiplier: Correct the calculated damage by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "31200"
    },
    "darkDiffenceRate": {
        "DisplayName": "Defence %: Holy",
        "Description": "Multiply the darkness defense by the set value",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "43200"
    },
    "darkDiffence": {
        "DisplayName": "Defence +: Holy",
        "Description": "Add or subtract the value set for darkness defense",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "46200"
    },
    "darkAttackRate": {
        "DisplayName": "Attack %: Holy",
        "Description": "Dark damage multiplier: Correct the calculated damage by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "34200"
    },
    "darkAttackPowerRate": {
        "DisplayName": "Power %: Holy",
        "Description": "Multiply the dark attack power by the set value",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "37200"
    },
    "darkAttackPower": {
        "DisplayName": "Damage +: Holy",
        "Description": "Add or subtract the value set for the dark attack power",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "40200"
    },
    "antiDarkSightRadius": {
        "DisplayName": "Full View in Darkness - Radius",
        "Description": "Radius of full view of darkness [m]. If you are within this distance, you will be able to see at normal distance even in the dark.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "SortID": "150000"
    },
    "antiDarkSightDmypolyId": {
        "DisplayName": "Full View in Darkness - Dummy Poly ID",
        "Description": "Damipoli ID (-1: Master) with full view of darkness. Create a full view area around this Damipoli",
        "Minimum": "-1",
        "Maximum": "99999",
        "SortID": "151000"
    },
    "conditionHpRate": {
        "DisplayName": "Trigger on HP Above %",
        "Description": "Activates only when you have HP above the specified value",
        "Minimum": "-1",
        "Maximum": "100",
        "SortID": "18500"
    },
    "consumeStaminaRate": {
        "DisplayName": "Weapon Stamina Consumption %",
        "Description": "Multiply by multiplying the consumption stamina value of the behavior parameter",
        "Minimum": "0",
        "Maximum": "99.99",
        "SortID": "55900"
    },
    "itemDropRate": {
        "DisplayName": "Item Discovery %",
        "Description": "The set value is added to [Item Drop Correction]\nRequires State Info 66",
        "Minimum": "-99.99",
        "Maximum": "99.99",
        "SortID": "72500"
    },
    "changePoisonResistPoint": {
        "DisplayName": "Aux Resist +: Poison",
        "Description": "Increase or decrease the state resistance value",
        "Minimum": "-99999",
        "Maximum": "99999",
        "SortID": "89000"
    },
    "changeDiseaseResistPoint": {
        "DisplayName": "Aux Resist +: Scarlet Rot",
        "Description": "Increase or decrease the state resistance value",
        "Minimum": "-99999",
        "Maximum": "99999",
        "SortID": "89100"
    },
    "changeBloodResistPoint": {
        "DisplayName": "Aux Resist +: Hemorrhage",
        "Description": "Increase or decrease the state resistance value",
        "Minimum": "-99999",
        "Maximum": "99999",
        "SortID": "89200"
    },
    "changeCurseResistPoint": {
        "DisplayName": "Aux Resist +: Blight",
        "Description": "Increase or decrease the state resistance value",
        "Minimum": "-99999",
        "Maximum": "99999",
        "SortID": "89300"
    },
    "changeFreezeResistPoint": {
        "DisplayName": "Aux Resist +: Frostbite",
        "Description": "Increase or decrease the state resistance value",
        "Minimum": "-99999",
        "Maximum": "99999",
        "SortID": "89400"
    },
    "slashAttackRate": {
        "DisplayName": "Attack %: Slash",
        "Description": "Slash damage ratio: Correct the calculated damage by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "32100"
    },
    "blowAttackRate": {
        "DisplayName": "Attack %: Strike",
        "Description": "Batter damage ratio: The calculated damage is corrected by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "32200"
    },
    "thrustAttackRate": {
        "DisplayName": "Attack %: Thrust",
        "Description": "Puncture damage ratio: Correct the calculated damage by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "32300"
    },
    "neutralAttackRate": {
        "DisplayName": "Attack %: Standard",
        "Description": "Non-attribute damage ratio: Correct the calculated damage by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "32400"
    },
    "slashAttackPowerRate": {
        "DisplayName": "Power %: Slash",
        "Description": "Multiply the slashing attack power by the set value",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "35100"
    },
    "blowAttackPowerRate": {
        "DisplayName": "Power %: Strike",
        "Description": "Multiply the hit attack power by the set value",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "35200"
    },
    "thrustAttackPowerRate": {
        "DisplayName": "Power %: Thrust",
        "Description": "Multiply the piercing attack power by the set value",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "35300"
    },
    "neutralAttackPowerRate": {
        "DisplayName": "Power %: Standard",
        "Description": "Multiply the non-attribute attack power by the set value",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "35400"
    },
    "slashAttackPower": {
        "DisplayName": "Damage +: Slash",
        "Description": "Add or subtract the value set for the slashing attack power",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "38100"
    },
    "blowAttackPower": {
        "DisplayName": "Damage +: Strike",
        "Description": "Add or subtract the value set for the batter attack power",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "38200"
    },
    "thrustAttackPower": {
        "DisplayName": "Damage +: Thrust",
        "Description": "Add or subtract the value set for the piercing attack power",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "38300"
    },
    "neutralAttackPower": {
        "DisplayName": "Damage +: Standard",
        "Description": "Add or subtract the value set for non-attribute attack power",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "38400"
    },
    "changeStrengthPoint": {
        "DisplayName": "Correction +: STR",
        "Description": "Add or subtract the correction value of the weapon",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "46600"
    },
    "changeAgilityPoint": {
        "DisplayName": "Correction +: DEX",
        "Description": "Add or subtract the correction value of the weapon",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "46610"
    },
    "changeMagicPoint": {
        "DisplayName": "Correction +: INT",
        "Description": "Add or subtract the correction value of the weapon",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "46700"
    },
    "changeFaithPoint": {
        "DisplayName": "Correction +: FTH",
        "Description": "Add or subtract the correction value of the weapon",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "46800"
    },
    "changeLuckPoint": {
        "DisplayName": "Correction +: ARC",
        "Description": "Add or subtract the correction value of the weapon",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "46900"
    },
    "recoverArtsPoint_Str": {
        "DisplayName": "Weapon Art Restore: STR",
        "Description": "Arts Point Restores strength",
        "Minimum": "0",
        "Maximum": "99",
        "SortID": "153000"
    },
    "recoverArtsPoint_Dex": {
        "DisplayName": "Weapon Art Restore: DEX",
        "Description": "Restores arts point workmanship",
        "Minimum": "0",
        "Maximum": "99",
        "SortID": "153100"
    },
    "recoverArtsPoint_Magic": {
        "DisplayName": "Weapon Art Restore: INT",
        "Description": "Restores arts point magic",
        "Minimum": "0",
        "Maximum": "99",
        "SortID": "153200"
    },
    "recoverArtsPoint_Miracle": {
        "DisplayName": "Weapon Art Restore: FTH",
        "Description": "Recover arts point miracles",
        "Minimum": "0",
        "Maximum": "99",
        "SortID": "153300"
    },
    "madnessDamageRate": {
        "DisplayName": "Damage Correction %: Madness",
        "Description": "Correction value used only for point damage and% damage of state change type [madness]",
        "SortID": "90300"
    },
    "isUseStatusAilmentAtkPowerCorrect": {
        "DisplayName": "Use Status Ailment Attack Power Correction",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "If ○, the abnormal state attack power multiplier correction of the attack para is applied.",
        "Maximum": "1",
        "SortID": "17500"
    },
    "isUseAtkParamAtkPowerCorrect": {
        "DisplayName": "Use AtkParam Attack Power Correction",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "If ○, the attack power multiplier correction of the attack para is applied.",
        "Maximum": "1",
        "SortID": "17550"
    },
    "dontDeleteOnDead": {
        "DisplayName": "Don't Delete on Dead",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "If it is ○, it will not be deleted even if the character dies. Mainly used for death effects.",
        "Maximum": "1",
        "SortID": "20500"
    },
    "disableFreeze": {
        "DisplayName": "Disable Frostbite",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "When this effect is applied, it will not be cold",
        "Maximum": "1",
        "SortID": "104060"
    },
    "isDisableNetSync": {
        "DisplayName": "Do not Sync in Multiplayer",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Do not synchronize with the net. It does not mean that you will be able to call it locally, but simply do not send it online. For example, a remote character does not activate locally, so nothing happens in that case.",
        "Maximum": "1",
        "SortID": "420000"
    },
    "shamanParamChange": {
        "DisplayName": "Affects Pyromancy",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Set whether or not it is effective against spells",
        "Maximum": "1",
        "SortID": "26100"
    },
    "isStopSearchedNotify": {
        "DisplayName": "Stop Platoon Notifications",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Whether to stop notifications targeting your army (used by EventMaker decisions and buddy platoons)",
        "Maximum": "1",
        "SortID": "75500"
    },
    "isCheckAboveShadowTest": {
        "DisplayName": "Apply outside of Rain Shield",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "If it is ○, it will not be applied when it is judged to be shielded (× is always applied)",
        "Maximum": "1",
        "SortID": "450000"
    },
    "addBehaviorJudgeId_add": {
        "DisplayName": "Add Behavior Judgment ID - Addition Value",
        "Description": "If the addition value of the action judgment ID is 0, the action is not switched and the action is stopped.",
        "Maximum": "999",
        "SortID": "114001"
    },
    "saReceiveDamageRate": {
        "DisplayName": "Absorption %: Poise Damage",
        "Description": "Multiplier for SA damage",
        "Minimum": "0",
        "Maximum": "100",
        "Increment": "0.1",
        "SortID": "90800"
    },
    "defPlayerDmgCorrectRate_Physics": {
        "DisplayName": "PVP Absorption %: Physical",
        "Description": "Damage correction for the damage value received from the player.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46300"
    },
    "defPlayerDmgCorrectRate_Magic": {
        "DisplayName": "PVP Absorption %: Magic",
        "Description": "Damage correction for the damage value received from the player.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46310"
    },
    "defPlayerDmgCorrectRate_Fire": {
        "DisplayName": "PVP Absorption %: Fire",
        "Description": "Damage correction for the damage value received from the player.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46320"
    },
    "defPlayerDmgCorrectRate_Thunder": {
        "DisplayName": "PVP Absorption %: Lightning",
        "Description": "Damage correction for the damage value received from the player.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46330"
    },
    "defPlayerDmgCorrectRate_Dark": {
        "DisplayName": "PVP Absorption %: Holy",
        "Description": "Damage correction for the damage value received from the player.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46340"
    },
    "defEnemyDmgCorrectRate_Physics": {
        "DisplayName": "Absorption %: Physical",
        "Description": "Damage correction for the damage value received from the enemy.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46350"
    },
    "defEnemyDmgCorrectRate_Magic": {
        "DisplayName": "Absorption %: Magic",
        "Description": "Damage correction for the damage value received from the enemy.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46360"
    },
    "defEnemyDmgCorrectRate_Fire": {
        "DisplayName": "Absorption %: Fire",
        "Description": "Damage correction for the damage value received from the enemy.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46370"
    },
    "defEnemyDmgCorrectRate_Thunder": {
        "DisplayName": "Absorption %: Lightning",
        "Description": "Damage correction for the damage value received from the enemy.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46380"
    },
    "defEnemyDmgCorrectRate_Dark": {
        "DisplayName": "Absorption %: Holy",
        "Description": "Damage correction for the damage value received from the enemy.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46390"
    },
    "defObjDmgCorrectRate": {
        "DisplayName": "Absorption %: Object",
        "Description": "Damage correction for the damage value received from OBJ.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46400"
    },
    "atkPlayerDmgCorrectRate_Physics": {
        "DisplayName": "PVP Damage %: Physical",
        "Description": "Damage correction for the damage value given to the player.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46290"
    },
    "atkPlayerDmgCorrectRate_Magic": {
        "DisplayName": "PVP Damage %: Magic",
        "Description": "Damage correction for the damage value given to the player.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46291"
    },
    "atkPlayerDmgCorrectRate_Fire": {
        "DisplayName": "PVP Damage %: Fire",
        "Description": "Damage correction for the damage value given to the player.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46292"
    },
    "atkPlayerDmgCorrectRate_Thunder": {
        "DisplayName": "PVP Damage %: Lightning",
        "Description": "Damage correction for the damage value given to the player.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46293"
    },
    "atkPlayerDmgCorrectRate_Dark": {
        "DisplayName": "PVP Damage %: Holy",
        "Description": "Damage correction for the damage value given to the player.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46294"
    },
    "atkEnemyDmgCorrectRate_Physics": {
        "DisplayName": "Damage %: Physical",
        "Description": "Damage correction for the damage value given to the enemy.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46295"
    },
    "atkEnemyDmgCorrectRate_Magic": {
        "DisplayName": "Damage %: Magic",
        "Description": "Damage correction for the damage value given to the enemy.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46296"
    },
    "atkEnemyDmgCorrectRate_Fire": {
        "DisplayName": "Damage %: Fire",
        "Description": "Damage correction for the damage value given to the enemy.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46297"
    },
    "atkEnemyDmgCorrectRate_Thunder": {
        "DisplayName": "Damage %: Lightning",
        "Description": "Damage correction for the damage value given to the enemy.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46298"
    },
    "atkEnemyDmgCorrectRate_Dark": {
        "DisplayName": "Damage %: Holy",
        "Description": "Damage correction for the damage value given to the enemy.",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "46299"
    },
    "registFreezeChangeRate": {
        "DisplayName": "Aux Resist %: Frostbite",
        "Description": "Multiply the cold resistance value by the set magnification",
        "Minimum": "-1",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "88400"
    },
    "invocationConditionsStateChange1": {
        "DisplayName": "Trigger on State Info [1]",
        "Enum": "SP_EFFECT_TYPE",
        "Description": "Trigger condition state change type 1",
        "Maximum": "60000",
        "SortID": "165100"
    },
    "invocationConditionsStateChange2": {
        "DisplayName": "Trigger on State Info [2]",
        "Enum": "SP_EFFECT_TYPE",
        "Description": "Trigger condition state change type 2",
        "Maximum": "60000",
        "SortID": "165200"
    },
    "invocationConditionsStateChange3": {
        "DisplayName": "Trigger on State Info [3]",
        "Enum": "SP_EFFECT_TYPE",
        "Description": "Trigger condition state change type 3",
        "Maximum": "60000",
        "SortID": "165300"
    },
    "hearingAiSoundLevel": {
        "DisplayName": "Listen - AI Sound Level",
        "Description": "Overwrite how good your ears are",
        "Minimum": "-1",
        "Maximum": "128",
        "SortID": "73850"
    },
    "chrProxyHeightRate": {
        "DisplayName": "Proxy Height Capsule %",
        "Description": "Magnification over the height of the character capsule",
        "DisplayFormat": "%0.2f",
        "Minimum": "0",
        "Maximum": "99",
        "SortID": "75700"
    },
    "addAwarePointCorrectValue_forMe": {
        "DisplayName": "Search Awareness - Addition Correction - Self",
        "Description": "Search side addition correction_viewer",
        "DisplayFormat": "%0.2f",
        "Minimum": "-9999",
        "Maximum": "9999",
        "Increment": "0.1",
        "SortID": "73500"
    },
    "addAwarePointCorrectValue_forTarget": {
        "DisplayName": "Search Awareness - Addition Correction - Target",
        "Description": "Searching degree addition correction _ side to be seen",
        "DisplayFormat": "%0.2f",
        "Minimum": "-9999",
        "Maximum": "9999",
        "Increment": "0.1",
        "SortID": "73510"
    },
    "sightSearchEnemyAdd": {
        "DisplayName": "Vision Search - Enemy Addition",
        "Description": "Correct the ease of finding with a real number",
        "DisplayFormat": "%0.2f",
        "Minimum": "-9999",
        "Maximum": "9999",
        "Increment": "0.1",
        "SortID": "73610"
    },
    "sightSearchAdd": {
        "DisplayName": "Vision Search Addition",
        "Description": "Correct the ease of finding with a real number",
        "DisplayFormat": "%0.2f",
        "Minimum": "-9999",
        "Maximum": "9999",
        "Increment": "0.1",
        "SortID": "73710"
    },
    "hearingSearchAdd": {
        "DisplayName": "Listen Search Addition",
        "Description": "Correct the hearing of AI sounds with real numbers",
        "DisplayFormat": "%0.2f",
        "Minimum": "-9999",
        "Maximum": "9999",
        "Increment": "0.1",
        "SortID": "73810"
    },
    "hearingSearchRate": {
        "DisplayName": "Listen Search Correction",
        "Description": "Correct the audibility of AI sound by magnification",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "73820"
    },
    "hearingSearchEnemyAdd": {
        "DisplayName": "Listen Search - Enemy Addition",
        "Description": "Correct the loudness of the emitted AI sound with a real number",
        "DisplayFormat": "%0.2f",
        "Minimum": "-9999",
        "Maximum": "9999",
        "Increment": "0.1",
        "SortID": "73910"
    },
    "value_Magnification": {
        "DisplayName": "Sale Price Correction %",
        "Description": "Sales price correction: Magnification",
        "Minimum": "1",
        "Maximum": "9999",
        "Increment": "0.1",
        "SortID": "200000"
    },
    "artsConsumptionRate": {
        "DisplayName": "Skill FP Consumption %",
        "Description": "Arts MP consumption multiplier [%]",
        "Minimum": "0",
        "Maximum": "990",
        "Increment": "0.001",
        "SortID": "300000"
    },
    "magicConsumptionRate": {
        "DisplayName": "Sorcery FP Consumption %",
        "Description": "Magic consumption MP multiplier [%]",
        "Minimum": "0",
        "Maximum": "990",
        "Increment": "0.001",
        "SortID": "300010"
    },
    "shamanConsumptionRate": {
        "DisplayName": "Pyromancy FP Consumption %",
        "Description": "Magic consumption MP multiplier [%]",
        "Minimum": "0",
        "Maximum": "990",
        "Increment": "0.001",
        "SortID": "300020"
    },
    "miracleConsumptionRate": {
        "DisplayName": "Incantation FP Consumption %",
        "Description": "Miracle consumption MP magnification [%]",
        "Minimum": "0",
        "Maximum": "990",
        "Increment": "0.001",
        "SortID": "300030"
    },
    "changeHpEstusFlaskRate": {
        "DisplayName": "HP Flask - HP Restore %",
        "Description": "Set what percentage of the maximum HP to add (or subtract) with one activation",
        "Minimum": "-100",
        "Maximum": "100",
        "SortID": "55200"
    },
    "changeHpEstusFlaskPoint": {
        "DisplayName": "HP Flask - HP Restore +",
        "Description": "Set how many points to add (or subtract) with one activation",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "55210"
    },
    "changeMpEstusFlaskRate": {
        "DisplayName": "FP Flask - FP Restore %",
        "Description": "Set what percentage of the maximum MP to add (or subtract) with one activation",
        "Minimum": "-100",
        "Maximum": "100",
        "SortID": "55220"
    },
    "changeMpEstusFlaskPoint": {
        "DisplayName": "FP Flask - FP Restore +",
        "Description": "Set how many points to add (or subtract) with one activation",
        "Minimum": "-9999",
        "Maximum": "9999",
        "SortID": "55230"
    },
    "changeHpEstusFlaskCorrectRate": {
        "DisplayName": "HP Flask - HP Restore Correction",
        "Description": "Correct the damage amount of the HP Est bottle",
        "Minimum": "0",
        "Maximum": "99999",
        "Increment": "0.001",
        "SortID": "55240"
    },
    "changeMpEstusFlaskCorrectRate": {
        "DisplayName": "FP Flask - HP Restore Correction",
        "Description": "Correct the damage amount of MP Est Bottle",
        "Minimum": "0",
        "Maximum": "99999",
        "Increment": "0.001",
        "SortID": "55250"
    },
    "applyIdOnGetSoul": {
        "DisplayName": "Kill SpEffect ID",
        "Description": "When the special effect of the state change type \"HP drain\" is enabled, when the enemy is defeated, the special effect ID set in the \"HP drain activation special effect\" of the same special effect is called (0: ignore).",
        "Minimum": "0",
        "Maximum": "2100000000",
        "SortID": "111030",
        "ParamRef1": "SpEffectParam"
    },
    "extendLifeRate": {
        "DisplayName": "Extend SpEffect Duration %",
        "Description": "Extension coefficient of state change type \"life extension\"",
        "Minimum": "0",
        "Maximum": "9900",
        "SortID": "105101"
    },
    "contractLifeRate": {
        "DisplayName": "Contract SpEffect Duration %",
        "Description": "Shortening coefficient of state change type \"life shortening\"",
        "Minimum": "0",
        "Maximum": "9900",
        "SortID": "105111"
    },
    "defObjectAttackPowerRate": {
        "DisplayName": "Object Damage %",
        "Description": "Corrects the attack power against the damage received from OBJ. (Not damage compensation)",
        "DisplayFormat": "%0.3f",
        "Minimum": "0",
        "Maximum": "99.999",
        "Increment": "0.001",
        "SortID": "46450"
    },
    "effectEndDeleteDecalGroupId": {
        "DisplayName": "Decal Deletion Group ID",
        "Description": "When the special effect disappears (lifetime / overwritten / erased ... etc.), the paint decal is deleted if the special effect of the same group ID is not applied.",
        "Minimum": "-1",
        "Maximum": "9999",
        "SortID": "410000"
    },
    "addLifeForceStatus": {
        "DisplayName": "Vigor",
        "Description": "Add value to growth status",
        "Minimum": "0",
        "Maximum": "99",
        "SortID": "401000"
    },
    "addWillpowerStatus": {
        "DisplayName": "Mind",
        "Description": "Add value to growth status",
        "Minimum": "0",
        "Maximum": "99",
        "SortID": "401100"
    },
    "addEndureStatus": {
        "DisplayName": "Endurance",
        "Description": "Add value to growth status",
        "Minimum": "0",
        "Maximum": "99",
        "SortID": "401200"
    },
    "addVitalityStatus": {
        "DisplayName": "Vitality",
        "Description": "Add value to growth status",
        "Minimum": "0",
        "Maximum": "99",
        "SortID": "401300"
    },
    "addStrengthStatus": {
        "DisplayName": "Strength",
        "Description": "Add value to growth status",
        "Minimum": "0",
        "Maximum": "99",
        "SortID": "401400"
    },
    "addDexterityStatus": {
        "DisplayName": "Dexterity",
        "Description": "Add value to growth status",
        "Minimum": "0",
        "Maximum": "99",
        "SortID": "401500"
    },
    "addMagicStatus": {
        "DisplayName": "Intelligence",
        "Description": "Add value to growth status",
        "Minimum": "0",
        "Maximum": "99",
        "SortID": "401600"
    },
    "addFaithStatus": {
        "DisplayName": "Faith",
        "Description": "Add value to growth status",
        "Minimum": "0",
        "Maximum": "99",
        "SortID": "401700"
    },
    "addLuckStatus": {
        "DisplayName": "Arcane",
        "Description": "Add value to growth status",
        "Minimum": "0",
        "Maximum": "99",
        "SortID": "401800"
    },
    "deleteCriteriaDamage": {
        "DisplayName": "Delete Criteria Damage",
        "Enum": "SP_EFFECT_PARAM_DELETE_DAMAGE_TYPE",
        "Description": "Reason for damage under the condition to remove special effects",
        "SortID": "18600"
    },
    "magicSubCategoryChange3": {
        "DisplayName": "Conditional Category [3]",
        "Enum": "ATK_SUB_CATEGORY",
        "Description": "Vs to subcategory parameter change 3",
        "SortID": "26320"
    },
    "spAttributeVariationValue": {
        "DisplayName": "SpEffect Attribute Variation Value",
        "Description": "This value is used to give variation to abnormal state SFX, SE, etc. in combination with the special attribute set for the special effect. SEQ16473",
        "Maximum": "99",
        "SortID": "27501"
    },
    "atkFlickPower": {
        "DisplayName": "Weapon Repel Power +",
        "Description": "Set a value that overwrites the repelling attack power",
        "SortID": "27505"
    },
    "wetConditionDepth": {
        "DisplayName": "Wet Condition Depth",
        "Enum": "SP_EFFECT_WET_CONDITION_DEPTH",
        "Description": "TimeAct Determines whether special effects are applied in combination with \"at what water level you get wet\"",
        "SortID": "460000"
    },
    "changeSaRecoveryVelocity": {
        "DisplayName": "Poise Recovery Time %",
        "Description": "Change the recovery speed of SA durability",
        "Minimum": "-99.99",
        "Maximum": "99.99",
        "Increment": "0.1",
        "SortID": "58300"
    },
    "regainRate": {
        "DisplayName": "Regain Correction %",
        "Description": "Regain magnification",
        "Minimum": "0",
        "Maximum": "99.99",
        "SortID": "162000"
    },
    "saAttackPowerRate": {
        "DisplayName": "Poise Damage %",
        "Description": "SA attack power multiplier",
        "Minimum": "0",
        "Maximum": "99",
        "Increment": "0.1",
        "SortID": "58400"
    },
    "sleepAttackPower": {
        "DisplayName": "Aux Inflict +: Sleep",
        "Description": "Numerical value to be added to the target's [Sleep tolerance] when hit",
        "Minimum": "-99999",
        "Maximum": "99999",
        "SortID": "66300"
    },
    "madnessAttackPower": {
        "DisplayName": "Aux Inflict +: Madness",
        "Description": "A number to be added to the target's [madness resistance value] when it hits",
        "Minimum": "-99999",
        "Maximum": "99999",
        "SortID": "66400"
    },
    "registSleepChangeRate": {
        "DisplayName": "Aux Resist %: Sleep",
        "Description": "Multiply the sleep tolerance value by the set magnification",
        "Minimum": "-1",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "88500"
    },
    "registMadnessChangeRate": {
        "DisplayName": "Aux Resist %: Madness",
        "Description": "Multiply the madness resistance value by the set multiplier",
        "Minimum": "-1",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "88600"
    },
    "changeSleepResistPoint": {
        "DisplayName": "Aux Resist +: Sleep",
        "Description": "Increase or decrease the state resistance value",
        "Minimum": "-99999",
        "Maximum": "99999",
        "SortID": "89500"
    },
    "changeMadnessResistPoint": {
        "DisplayName": "Aux Resist +: Madness",
        "Description": "Increase or decrease the state resistance value",
        "Minimum": "-99999",
        "Maximum": "99999",
        "SortID": "89600"
    },
    "sleepDamageRate": {
        "DisplayName": "Damage Correction %: Sleep",
        "Description": "Point damage of state change type [Sleep], correction value used only when% damage",
        "SortID": "90200"
    },
    "applyPartsGroup": {
        "DisplayName": "Apply Parts Group",
        "Enum": "SP_EFFECT_APPLY_PARTS_GROUP",
        "Description": "The effect is limited by the part where the attack hits. Only defensive items in damage calculation are subject to restriction",
        "Maximum": "31",
        "SortID": "26400"
    },
    "clearTarget": {
        "DisplayName": "Clear Target",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "Does not recognize the target while the special effect is applied (excluding the riding target)",
        "Maximum": "1",
        "SortID": "72710"
    },
    "fakeTargetIgnoreAjin": {
        "DisplayName": "Ignore Fake Target - Human",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "You will not be caught by the fake target of the subhuman system that occurred",
        "Maximum": "1",
        "SortID": "96602"
    },
    "fakeTargetIgnoreMirageArts": {
        "DisplayName": "Ignore Fake Target - Phantom",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "You will not be caught by the fake target of the phantom arts system that occurred",
        "Maximum": "1",
        "SortID": "96604"
    },
    "requestForceJoinBlackSOS_B": {
        "DisplayName": "Force Invasion",
        "Enum": "SP_EFFECT_BOOL",
        "Description": "If checked, issue an intrusion_B request when activated",
        "Maximum": "1",
        "SortID": "78100"
    },
    "unk353_4": {
        "DisplayName": "Blank",
        "Enum": "SP_EFFECT_BOOL",
        "Maximum": "1"
    },
    "pad2[1]": {
        "DisplayName": "pad",
        "SortID": "470001"
    },
    "changeSuperArmorPoint": {
        "DisplayName": "Poise +",
        "Description": "Value to add to the super armor value",
        "Minimum": "-9999",
        "Maximum": "9999",
        "Increment": "0.1",
        "SortID": "58100"
    },
    "changeSaPoint": {
        "DisplayName": "Apply Poise Damage +",
        "Description": "Set how many points to subtract (or add) with one activation",
        "Minimum": "-9999",
        "Maximum": "9999",
        "Increment": "0.1",
        "SortID": "58200"
    },
    "hugeEnemyPickupHeightOverwrite": {
        "DisplayName": "Giant Enemy Height Overwrite",
        "Description": "Giant enemy lift height overwrite [m]",
        "Minimum": "0",
        "Maximum": "99.9",
        "Increment": "0.1",
        "SortID": "470000"
    },
    "poisonDefDamageRate": {
        "DisplayName": "Aux Damage Correction %: Poison",
        "Description": "Poison resistance damage ratio: Correct the calculated damage by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "31300"
    },
    "diseaseDefDamageRate": {
        "DisplayName": "Aux Damage Correction %: Scarlet Rot",
        "Description": "Epidemic resistance damage ratio: Correct the calculated damage by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "31400"
    },
    "bloodDefDamageRate": {
        "DisplayName": "Aux Damage Correction %: Hemorrhage",
        "Description": "Bleeding resistance damage ratio: Correct the calculated damage by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "31500"
    },
    "curseDefDamageRate": {
        "DisplayName": "Aux Damage Correction %: Blight",
        "Description": "Curse resistance damage multiplier: Corrects the calculated damage by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "31600"
    },
    "freezeDefDamageRate": {
        "DisplayName": "Aux Damage Correction %: Frostbite",
        "Description": "Cold resistance damage ratio: Correct the calculated damage by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "31700"
    },
    "sleepDefDamageRate": {
        "DisplayName": "Aux Damage Correction %: Sleep",
        "Description": "Sleep resistance damage ratio: Correct the calculated damage by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "31800"
    },
    "madnessDefDamageRate": {
        "DisplayName": "Aux Damage Correction %: Madness",
        "Description": "Madness resistance damage multiplier: Correct the calculated damage by XX times. 1 is normal.",
        "Minimum": "-99",
        "Maximum": "99",
        "Increment": "0.001",
        "SortID": "31900"
    },
    "overwrite_maxBackhomeDist": {
        "DisplayName": "Override - Max Back Home Distance",
        "Description": "Distance to go home no matter what [m] _ overwrite",
        "SortID": "74000"
    },
    "overwrite_backhomeDist": {
        "DisplayName": "Override - Back Home Distance",
        "Description": "Distance to return home while fighting [m] _ Overwrite",
        "SortID": "74010"
    },
    "overwrite_backhomeBattleDist": {
        "DisplayName": "Override - Back Home Battle Distance",
        "Description": "Distance to give up and fight to return to the nest [m] _ overwrite",
        "SortID": "74020"
    },
    "overwrite_BackHome_LookTargetDist": {
        "DisplayName": "Override - Look Target Distance",
        "Description": "When returning home: Distance to see the target [m] _ Overwrite",
        "SortID": "74030"
    },
    "goodsConsumptionRate": {
        "DisplayName": "Item FP Consumption %",
        "Description": "Item consumption MP multiplier",
        "Minimum": "0",
        "Maximum": "990",
        "Increment": "0.001",
        "SortID": "300040"
    },
    "pad3[8]": {
        "DisplayName": "pad",
        "DisplayFormat": "%f",
        "EditFlags": "Wrap",
        "SortID": "470002"
    }
}


from typing import Any

from erdb.typing.models.correction_attack import CorrectionAttack, Correction, Override, Ratio
from erdb.typing.params import ParamRow
from erdb.typing.enums import ItemIDFlag
from erdb.typing.api_version import ApiVersion
from erdb.table._retrievers import ParamDictRetriever, RetrieverData
from erdb.table._common import TableSpecContext


_DAMAGE_TYPE: dict[str, str] = {
    "physical": "Physics",
    "magic": "Magic",
    "fire": "Fire",
    "lightning": "Thunder",
    "holy": "Dark",
}

_ATTRIBUTE: dict[str, str] = {
    "strength": "Strength",
    "dexterity": "Dexterity",
    "intelligence": "Magic",
    "faith": "Faith",
    "arcane": "Luck",
}

def _get_attributes(row: ParamRow, damage_type: str, cls: Any) -> Any:
    def get_field(attribute: str, damage_type: str) -> int:
        field = cls.get_property(_ATTRIBUTE[attribute], _DAMAGE_TYPE[damage_type])

        return {
            Correction: row[field].as_bool,
            Override: value / 100.0 if (value := row[field].get_float()) else None,
            Ratio: row[field].as_float / 100.0,
        }[cls]

    ret = {attrib: get_field(attrib, damage_type) for attrib in _ATTRIBUTE.keys()}
    ret = {k: v for k, v in ret.items() if v is not None}
    return cls.get_field_type()(**ret)

def _get_damage_types(row: ParamRow, cls: Any) -> Any:
    data = {damage: _get_attributes(row, damage, cls) for damage in _DAMAGE_TYPE.keys()}
    return cls(**data)

class CorrectionAttackTableSpec(TableSpecContext):
    model = {
        ApiVersion.VER_1: CorrectionAttack,
    }

    main_param_retriever = ParamDictRetriever("AttackElementCorrectParam", ItemIDFlag.NON_EQUIPABBLE)

    @classmethod # override
    def get_pk(cls, data: RetrieverData, row: ParamRow) -> str:
        return str(row.index)

    @classmethod
    def make_object(cls, api: ApiVersion, data: RetrieverData, row: ParamRow):
        return CorrectionAttack(
            correction=_get_damage_types(row, Correction),
            override=_get_damage_types(row, Override),
            ratio=_get_damage_types(row, Ratio),
        )


from pydantic import Field, PositiveInt, NonNegativeInt
from pydantic.dataclasses import dataclass

from erdb.typing.models import dt_config, NonEmptyStr
from erdb.typing.models.item import Item
from erdb.typing.enums import SpiritAshUpgradeMaterial


@dataclass(config=dt_config())
class SpiritAshUserData:
    summon_quantity: PositiveInt = Field(1,
        description="The number of spririts this Spirit Ash summons.",
        example=1,
    )
    abilities: list[NonEmptyStr] = Field([],
        description="Strenghts of the Spirit Ash. Each briefly described with minimum amount of words.",
        example=[
            "Highly mobile",
            "Casts Blade of Death",
            "Jump and charged attacks can knock enemies down",
        ],
    )

@dataclass(config=dt_config())
class SpiritAsh(SpiritAshUserData, Item):
    summon_name: str = Field(...,
        description="Specific name of the summoned spirit(s). Rarely differs from Item name.",
        min_length=1,
        example="Black Knife Tiche"
    )
    fp_cost: NonNegativeInt = Field(...,
        description="Cost of FP to summon, 0 if not applicable.",
        example=132,
    )
    hp_cost: NonNegativeInt = Field(...,
        description="Cost of HP to summon, 0 if not applicable.",
        example=0,
    )
    upgrade_material: SpiritAshUpgradeMaterial = Field(...,
        description="Glovewort the Spirit Ash upgrades with.",
        example=SpiritAshUpgradeMaterial.GHOST_GLOVEWORT,
    )
    upgrade_costs: list[NonNegativeInt] = Field(...,
        description="Array of Rune costs to upgrade to each level, +1 starting at position 0.",
        min_items=10, max_items=10,
        example=[2000, 3200, 4400, 5600, 6800, 8000, 9200, 10400, 11600, 14000]
    )


import json
import shutil
import string
import requests
import subprocess as proc
import xml.etree.ElementTree as xmltree
from PIL import Image
from io import BytesIO
from csv import DictReader
from zipfile import ZipFile, ZIP_DEFLATED
from itertools import chain, islice
from time import sleep
from hashlib import md5
from pathlib import Path
from typing import NamedTuple, Self

from erdb.table import Table
from erdb.loaders import PKG_DATA_PATH
from erdb.utils.common import get_filename
from erdb.typing.game_version import GameVersion, GameVersionInstance
from erdb.typing.params import ParamRow
from erdb.typing.enums import ItemIDFlag


def _prepare_writable_path(path: Path, default_filename: str) -> Path:
    (path if path.suffix == "" else path.parent).mkdir(parents=True, exist_ok=True)
    return path / default_filename if path.is_dir() else path

def _is_empty(d: Path) -> bool:
    assert d.is_dir(), f"{d} is not a directory."
    return not any(d.iterdir())

def _load_manifest():
    with open(PKG_DATA_PATH / "gamedata" / "manifest.json") as f:
        return json.load(f)

def _process_cache(*paths: Path, keep_cache: bool):
    def get_size(d: Path) -> str:
        assert d.is_dir(), f"{d} is not a directory."
        size = sum(p.stat().st_size for p in d.rglob("*"))
        return f"{size:,}"

    if keep_cache:
        print("Keeping unpacked files due to --keep-cache option:", flush=True)
        for p in paths: print("*", p, get_size(p), "bytes", flush=True)

    else:
        print("Removing unpacked files (--keep-cache not provided)...", flush=True)
        for p in paths: shutil.rmtree(p, ignore_errors=True)

def _get_app_version(game_exe: Path) -> list[int]:
    try:
        import win32api as w
    except ModuleNotFoundError:
        assert False, "Module \"pywin32\" is not installed, please install it manually"

    info = w.GetFileVersionInfo(str(game_exe), "\\")
    ms = info["FileVersionMS"] # type: ignore
    ls = info["FileVersionLS"] # type: ignore
    major, minor, patch = w.HIWORD(ms), w.LOWORD(ms), w.HIWORD(ls)

    return [int(major), int(minor), int(patch)]

def _get_effective_version(game_dir: Path) -> GameVersion:
    app_version = _get_app_version(game_dir / "eldenring.exe")
    version_instance = GameVersionInstance.construct(app_version, game_dir / "regulation_version.txt")
    print(f"Detected versions: {version_instance}.", flush=True)
    return version_instance.effective

class _File(NamedTuple):
    path: Path
    verifier: str
    verifier_is_md5: bool

    @classmethod
    def from_manifest(cls, tool_path: Path, name: str, verifier: str, **kwargs) -> Self:
        verifier = verifier.format(**kwargs)
        verifier_is_md5 = all(c in string.hexdigits for c in verifier)
        return cls(tool_path / Path(name), verifier, verifier_is_md5)

    def __str__(self) -> str:
        return str(self.path)

class _Command(NamedTuple):
    args: list[str]

    def run(self, cwd: Path):
        print(f"> [{cwd}]", *self.args, flush=True)

        p = proc.Popen(self.args, cwd=cwd, stdout=proc.PIPE, stderr=proc.STDOUT, text=True, shell=True, bufsize=1)
        assert p.stdout is not None

        for line in p.stdout:
            print(line.strip(), flush=True)

        while p.poll() is None:
            sleep(0.2)

        assert p.returncode == 0, f"Command failed with return code {p.returncode}."

    @classmethod
    def make(cls, args: list[str], **kwargs) -> Self:
        return cls([arg.format(**kwargs) for arg in args])

class _Tool(NamedTuple):
    name: str
    url: str
    path: Path
    files: list[_File]
    commands: list[_Command]

    def check_files(self, ignore_checksum: bool):
        print(f"Validating files in \"{self.path}\"...", flush=True)

        for f in self.files:
            print("*", f, flush=True)

            if f.verifier_is_md5:
                assert f.path.exists(), f"{f} does not exist."

                if md5(f.path.read_bytes()).hexdigest() != f.verifier:
                    if ignore_checksum:
                        print(f"WARNING: Checksum for {f} mismatched, but is being ignored.", flush=True)
                    else:
                        assert False, f"Mismatched checksum of {f}."

            else:
                print(f"Copying {f.verifier} to {f.path}...")
                shutil.copy(f.verifier, f.path)

        print(f"{self.name} files validated successfully!", flush=True)

    def run_commands(self):
        print(f"\nRunning {len(self.commands)} tool command(s).", flush=True)
        for cmd in self.commands:
            cmd.run(self.path)

    def run_command(self, *args: str):
        _Command.make([*args]).run(self.path)

    @classmethod
    def from_manifest(cls, game_dir: Path, name: str, data: dict, skip_commands: bool = False) -> Self:
        path = PKG_DATA_PATH / "thirdparty" / name

        if not path.exists() or _is_empty(path):
            print(name, "is not found and will be downloaded...", flush=True)
            _Tool.download_and_unzip(data["url"], path)

        files = [_File.from_manifest(path, name, verifier, game_dir=game_dir) for name, verifier in data["files"].items()]

        commands = \
            [] if skip_commands else \
            [_Command.make(args, game_dir=game_dir) for args in data["cmds"]]

        return cls(name, data["url"], path, files, commands)

    @classmethod
    def load_all(cls, game_dir: Path, manifest):
        return [cls.from_manifest(game_dir, name, data) for name, data in manifest["tools"].items()]

    @classmethod
    def load_custom(cls, game_dir: Path, manifest, *names: str):
        return [cls.from_manifest(game_dir, name, manifest["tools"][name], skip_commands=True) for name in names]

    @staticmethod
    def download_and_unzip(url: str, destination: Path):
        print("Downloading", url, flush=True)

        resp = requests.get(url)

        z = ZipFile(BytesIO(resp.content))
        top_level_names = {name.split("/")[0] for name in z.namelist()}

        # each tool should have <ToolName>/<files> as contents of its zip
        assert len(top_level_names) == 1

        main_dir = next(iter(top_level_names))

        z.extractall(destination.parent)
        shutil.move(destination.parent / main_dir, destination)

        print("Extracted tool files to:", destination, flush=True)

class _MapTile(NamedTuple):
    Coords = tuple[int, int]
    Code = int
    Masks = dict[Coords, Code]

    path: Path
    underground: bool
    lod: int
    x: int
    y: int
    code: Code

    @property
    def stem(self) -> str:
        ext = "".join(self.path.suffixes)
        return self.path.name.removesuffix(ext)

    @property
    def coords(self) -> Coords:
        return (self.x, self.y)

    @property
    def is_out_of_bounds(self) -> bool:
        """
        Tiles with this bit are always out of bounds, filtering them
        will allow us to trim more unused tiles at edges of the map.
        """
        return bool(self.code & (1 << 14))

    @classmethod
    def from_path(cls, path: Path):
        # split until first dot, not necessairly extension in case of ".tpf.dcx"
        filename, _ = path.name.split(".", 1)

        parts = filename.split("_")
        assert len(parts) == 7, f"Invalid map tile filename: {path}"

        return cls(path, parts[2] == "M01", int(parts[3][1]), int(parts[4]), int(parts[5]), int(parts[6], 16))

    @staticmethod
    def glob(path: Path, extension: str, lod: int, underground: bool, masks: Masks) -> list["_MapTile"]:
        glob_filter = f"*MENU_MapTile_M0{1 if underground else 0}_L{lod}*{extension}"
        tiles = (_MapTile.from_path(p) for p in path.rglob(glob_filter))

        # filter the valid, fully-unlocked tile variants based on their bitmask
        return [t for t in tiles if masks.get(t.coords) == t.code and not t.is_out_of_bounds]

def source_gamedata(game_dir: Path, ignore_checksum: bool, version: GameVersion | None = None):
    # TODO: check if `game_dir` is UXM-unpacked (backup_ directory?)
    #       preferably do unpacking via custom CLI tool mimicing UXM

    manifest = _load_manifest()
    tools = _Tool.load_all(game_dir, manifest)

    for tool in tools:
        print(f"\nUsing tool \"{tool.name}\", acquired from \"{tool.url}\".", flush=True)
        tool.check_files(ignore_checksum)
        tool.run_commands()

    if version is None:
        version = _get_effective_version(game_dir)

    print(f"Effective version: {version}.", flush=True)

    archive = PKG_DATA_PATH / "gamedata" / f"{version}.zip"
    assert not archive.exists(), f"{archive} exists and is not empty."

    print(f"Adding files to archive...", flush=True)
    with ZipFile(archive, "w", compression=ZIP_DEFLATED) as z:
        for filename, metadata in manifest["gamedata"].items():
            location = metadata["location"].format(game_dir=game_dir)
            z.write(Path(location) / filename, arcname=filename)

    print(f"Sourcing version {version} complete!", flush=True)

def _assemble_map(tiles: list[_MapTile], out: Path | None = None):
    if len(tiles) == 0:
        print(f"WARNING: No tiles found with specified options.", flush=True)
        return

    low_x,  low_y  = min(t.x for t in tiles), min(t.y for t in tiles)
    high_x, high_y = max(t.x for t in tiles), max(t.y for t in tiles)

    # X and Y are number of tiles, add 1 for 0th row and col
    X, Y = high_x - low_x + 1, high_y - low_y + 1

    lod = tiles[0].lod
    level = "underground" if tiles[0].underground else "overworld"

    print(f"Generating {X}x{Y} {level} map file with lod {lod}...", flush=True)

    expected_size = 256
    worldmap = Image.new("RGB", (X * expected_size, Y * expected_size))

    for tile in tiles:
        with Image.open(tile.path) as img:
            assert expected_size == img.width == img.height
            x, y = tile.x - low_x, Y - tile.y + low_y - 1
            worldmap.paste(img, (x * expected_size, y * expected_size))

    if out is None:
        worldmap.show()
        return

    out = _prepare_writable_path(out, f"er_map_{level}_l{lod}.jpeg")
    print(f"Writing to {out}...", flush=True)

    worldmap.save(out)

def _parse_tile_masks(mask_dir: Path, lod: int = 0, underground: bool = False) -> _MapTile.Masks:
    masks: _MapTile.Masks = dict()

    mtmsk = f"MENU_MapTile_M0{1 if underground else 0}.mtmsk"
    tree = xmltree.parse(mask_dir / mtmsk)

    for entry in tree.getroot().findall(".//MapTileMask"):
        index = int(str(entry.attrib.get("id")))
        this_lod, coords = divmod(index, 10000)

        if this_lod == lod:
            x, y = divmod(coords, 100)
            masks[(x, y)] = int(str(entry.attrib.get("mask")))

    return masks

def _unpack_missing_dds(yabber: _Tool, dds_files: list[Path]):
    def chunks(files_iterable, size: int):
        it = iter(files_iterable)
        for first in it:
            yield chain([first], islice(it, size - 1))

    files = [f for f in dds_files if not f.exists()]
    files = [dcx for f in files if (dcx := f.parent.parent / f.with_suffix(".tpf.dcx").name).exists()]

    for files_chunk in chunks(files, 20):
        yabber.run_command("Yabber.exe", *map(str, files_chunk))

def source_map(game_dir: Path, out: Path | None = None, lod: int = 0, underground: bool = False, ignore_checksum: bool = False, keep_cache: bool = False):
    manifest = _load_manifest()

    yabber, = _Tool.load_custom(game_dir, manifest, "Yabber")
    yabber.check_files(ignore_checksum)

    tile_dir = game_dir / "menu" / "71_maptile-tpfbhd" / "71_MapTile"
    mask_dir = game_dir / "menu" / "71_maptile-mtmskbnd-dcx" / "GR" / "data" / "INTERROOT_win64" / "menu" / "ScaleForm" / "maptile" / "mask"

    try:

        if not tile_dir.is_dir() or _is_empty(tile_dir):
            yabber.run_command("Yabber.exe", f"{game_dir}/menu/71_maptile.tpfbhd")

        if not mask_dir.is_dir() or _is_empty(mask_dir):
            yabber.run_command("Yabber.exe", f"{game_dir}/menu/71_maptile.mtmskbnd.dcx")

        masks = _parse_tile_masks(mask_dir, lod, underground)

        dcx_tiles = _MapTile.glob(tile_dir, ".tpf.dcx", lod, underground, masks)
        dds_files = [t.path.parent / f"{t.stem}-tpf-dcx" / f"{t.stem}.dds" for t in dcx_tiles]

        _unpack_missing_dds(yabber, dds_files)

        _assemble_map(_MapTile.glob(tile_dir, ".dds", lod, underground, masks), out)

    except: raise
    finally: _process_cache(tile_dir, mask_dir, keep_cache=keep_cache)

def source_icons(game_dir: Path, tables: list[Table], size: int, file_format: str, desination: Path, ignore_checksum: bool = False, keep_cache: bool = False):
    assert 1 <= size <= 1024, f"Invalid size: {size}"

    def readr(tb: Table):
        """
        Need the make sure the item name is used from the very first occurence
        of the icon ID, otherwise icon IDs will be assigned to a fully upgraded
        version of the item since upgrades share the icon. Reverse reading the
        param is easier for dict comprehension, because instead of checking if a
        key exist, we can keep overriding and keep the code simpler.
        """
        with open(game_dir / "param" / "gameparam" / f"{tb.param_name}.csv", mode="r") as f:
            reader = DictReader(f, delimiter=";")
            rows = {int(row["Row ID"]): row for row in reader}

            for index in reversed(rows.keys()):
                yield ParamRow.make(rows[index], ItemIDFlag.DISABLE_CHECK)

    def get_names(tb: Table):
        assert "names" in tb.spec.msg_retrievers, f"Table {tb} must specify a \"names\" msg retriever to write icons."

        names_file = names_dir / "GR" / "data" / "INTERROOT_win64" / "msg" / "engUS" / (tb.spec.msg_retrievers["names"].file_name + ".fmg")
        yabber.run_command("Yabber.exe", str(names_file))

        tree = xmltree.parse(names_file.with_suffix(".fmg.xml"))
        entries = tree.findall(".//text")

        return {int(str(e.get("id"))) : str(e.text) for e in entries}

    def is_valid_row(tb: Table, row: ParamRow) -> bool:
        return all(pred(row) for pred in tb.spec.predicates) and row.index in tb.spec.main_param_retriever

    def get_icon_id(row: ParamRow) -> int:
        return row["iconId"].as_int if "iconId" in row else row["iconIdM"].as_int

    def get_icon_dds(loc: Path, icon_id: int) -> Path:
        stem = f"MENU_Knowledge_{icon_id:05}"
        return loc / "00_Solo" / f"{stem}-tpf-dcx" / f"{stem}.dds"

    manifest = _load_manifest()

    yabber, er_exporter = _Tool.load_custom(game_dir, manifest, "Yabber", "ERExporter.Param")
    yabber.check_files(ignore_checksum)
    er_exporter.check_files(ignore_checksum)

    table_dir = game_dir / "param" / "gameparam"
    names_dir = game_dir / "msg" / "engus" / "item-msgbnd-dcx"
    icon_dir = game_dir / "menu" / "hi" / "00_solo-tpfbhd"
    desination.mkdir(parents=True, exist_ok=True)

    try:

        if not table_dir.is_dir() or _is_empty(table_dir):
            er_exporter.run_command("ERExporter.Param.exe", f"{game_dir}/regulation.bin")

        param_files = (game_dir / "param" / "gameparam" / f"{tb.param_name}.param" for tb in tables)
        param_files = [f for f in param_files if not f.with_suffix(".csv").exists()]

        if len(param_files) > 0:
            er_exporter.run_command("ERExporter.Param.exe", *map(str, param_files))

        if not names_dir.is_dir() or _is_empty(names_dir):
            yabber.run_command("Yabber.exe", str(names_dir.parent / "item.msgbnd.dcx"))

        if not icon_dir.is_dir() or _is_empty(icon_dir):
            yabber.run_command("Yabber.exe", str(icon_dir.parent / "00_solo.tpfbhd"))

        for tb in tables:
            names = get_names(tb)
            dest = desination / str(tb)
            dest.mkdir()

            print(f"Exporting to {dest}...", flush=True)

            id_to_name = {get_icon_id(row): get_filename(names[row.index]) for row in readr(tb) if is_valid_row(tb, row)}
            dds_files  = [get_icon_dds(icon_dir, icon_id) for icon_id in id_to_name.keys()]

            _unpack_missing_dds(yabber, dds_files)

            cur, count = 0, len(id_to_name)

            for icon_id, name in id_to_name.items():
                dds = get_icon_dds(icon_dir, icon_id)

                if not dds.exists():
                    print(f"[{(cur := cur + 1)}/{count}] Skipping missing icon for {name}: {dds}", flush=True)
                    continue

                with Image.open(dds) as img:
                    out = dest / file_format.format(icon_id=icon_id, name=name, table=tb)

                    print(f"[{(cur := cur + 1)}/{count}] {out}", flush=True)
                    img.resize((size, size)).save(out)

    except: raise
    finally: _process_cache(table_dir, names_dir, icon_dir, keep_cache=keep_cache)



from pydantic import Field
from pydantic.dataclasses import dataclass

from erdb.typing.models import dt_config


@dataclass(config=dt_config())
class Damage:
    physical: int | None = None
    magic: int | None = None
    fire: int | None = None
    lightning: int | None = None
    holy: int | None = None
    stamina: int | None = None

@dataclass(config=dt_config())
class StatRequirements:
    strength: int | None = Field(None, ge=0, le=99)
    dexterity: int | None = Field(None, ge=0, le=99)
    intelligence: int | None = Field(None, ge=0, le=99)
    faith: int | None = Field(None, ge=0, le=99)
    arcane: int | None = Field(None, ge=0, le=99)


import re
from functools import total_ordering
from pathlib import Path
from typing import Any, Generator, NamedTuple, Self


@total_ordering
class GameVersion(NamedTuple):
    major: str
    minor: str
    patch: str
    nums: list[int]

    @classmethod
    def from_nums(cls, major_int: int, minor_int: int, patch_int: int) -> Self:
        major = str(major_int)
        minor = f"{minor_int:02}"
        patch = str(patch_int)

        assert len(major) == 1
        assert len(minor) == 2
        assert len(patch) == 1

        return cls(major, minor, patch, [major_int, minor_int, patch_int])

    @classmethod
    def from_string(cls, version: str) -> Self:
        parts = version.split(".")
        assert len(parts) == 3, "Invalid version string given"
        assert len(parts[1]) >= 2, "Minor part must be at least 2 digits"

        nums = [int(parts[0]), int(parts[1]), int(parts[2])]
        return cls(parts[0], parts[1], parts[2], nums)

    @classmethod
    def from_any(cls, obj: Any) -> Self:
        if isinstance(obj, GameVersion):
            return obj

        if isinstance(obj, str):
            return GameVersion.from_string(obj)

        if isinstance(obj, Path):
            with open(obj, "r") as f:
                data = f.read()
            return GameVersion.from_string(data)

        if isinstance(obj, list):
            assert len(obj) == 3 and all(isinstance(elem, int) for elem in obj)
            return GameVersion.from_nums(obj[0], obj[1], obj[2])

        assert False, "Cannot parse GameVersion"

    @classmethod
    def min(cls) -> Self:
        return cls("0", "00", "0", [0, 0, 0])

    @classmethod
    def max(cls) -> Self:
        return cls("99999", "99999", "99999", [99999, 99999, 99999])

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.patch}"

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, GameVersion):
            return False
        return self.major == __o.major and self.minor == __o.minor and self.patch == __o.patch

    def __lt__(self, __o: "GameVersion") -> bool:
        assert isinstance(__o, GameVersion)
        for this_num, other_num in zip(self.nums, __o.nums):
            if this_num != other_num:
                return this_num < other_num
        return False

class GameVersionRange(NamedTuple):
    begin: GameVersion # including
    end: GameVersion   # excluding
    only: bool=False   # only `begin`

    def iterate(self, versions: list[GameVersion]) -> Generator[GameVersion, None, None]:
        for version in versions:
            if version in self:
                yield version

    @classmethod
    def from_version(cls, version: GameVersion) -> Self:
        return cls(version, GameVersion.max(), only=True)

    @classmethod
    def from_string(cls, string: str) -> Self:
        def _ver(match: re.Match[str]) -> GameVersion:
            return GameVersion.from_string(match.group(1))

        if string == "any version":
            return cls(GameVersion.min(), GameVersion.max())

        if search_only := re.search(r"only (\d+\.\d\d+\.\d+)", string):
            return cls(_ver(search_only), GameVersion.max(), only=True)

        search_begin = re.search(r"from (\d+\.\d\d+\.\d+)", string)
        search_end = re.search(r"until (\d+\.\d\d+\.\d+)", string)
        assert search_begin or search_end, "Invalid version range string"

        begin = _ver(search_begin) if search_begin else GameVersion.min()
        end = _ver(search_end) if search_end else GameVersion.max()

        assert begin < end, "Invalid version range string"
        return cls(begin, end)

    def __contains__(self, version: GameVersion) -> bool:
        assert isinstance(version, GameVersion)
        return self.begin == version if self.only \
            else self.begin <= version and version < self.end

class GameVersionInstance(NamedTuple):
    application: GameVersion
    regulation: GameVersion

    @property
    def effective(self) -> GameVersion:
        return max(self.application, self.regulation)

    @classmethod
    def construct(cls, application: Any, regulation: Any) -> Self:
        return cls(GameVersion.from_any(application), GameVersion.from_any(regulation))

    def __str__(self) -> str:
        return f"(app: {self.application}, regulation: {self.regulation})"



from pydantic import Field, PositiveInt, NonNegativeInt
from pydantic.dataclasses import dataclass

from erdb.typing.models import dt_config, NonEmptyStr
from erdb.typing.models.location import LocationDetail
from erdb.typing.enums import GoodsRarity


@dataclass(config=dt_config())
class Item:
    full_hex_id: str = Field(...,
        description="Full ID of the Item in capital hexadecimal form. IDs differ per affinity or upgrade level.",
        regex=r"^[0-9A-F]+$",
        min_length=8, max_length=8,
        example="400000BE",
    )
    id: PositiveInt = Field(...,
        description="ID of the Item in its individual class. IDs differ per affinity or upgrade level.",
        example=190,
    )
    name: str = Field(...,
        description="Name of the Item.",
        min_length=1,
        example="Rune Arc",
    )
    summary: str = Field(...,
        description="Short description of the Item.",
        min_length=1,
        example="Grants the blessing of an equipped Great Rune upon use",
    )
    description: list[str] = Field(...,
        description="Array of lines of the in-game description, each element is separated by a new line. A line may contain multiple sentences, or be empty if in between paragraphs.",
        min_items=1,
        example=[
            "A shard of the shattered Elden Ring.",
            "Grants the blessing of an equipped Great Rune upon use.",
            "",
            "Even if no Great Rune is equipped, it will slightly increase maximum HP upon use.",
            "",
            "The lower arc of the Elden Ring is held to be the basin in which its blessings pool. Perhaps this shard originates from that very arc.",
        ],
    )
    is_tradable: bool = Field(...,
        description="Specifies whether the Item is visible to other players if dropped.",
    )
    price_sold: NonNegativeInt = Field(...,
        description="The amount of Runes the Item is sold for, 0 if not applicabe.",
        example=200,
    )
    rarity: GoodsRarity = Field(...,
        description="Rarity of the Item.",
        example=GoodsRarity.COMMON,
    )
    icon: NonNegativeInt = Field(...,
        description="ID of the icon which can be shared across many items. Icons can be sourced from the game files using ERDB.",
        example=584,
    )
    max_held: NonNegativeInt = Field(...,
        description="The maximum amount of the Item that a player can have on them.",
        example=99,
    )
    max_stored: NonNegativeInt = Field(...,
        description="The maximum amount of the Item that can be stored in the sort chest.",
        example=600,
    )
    locations: list[LocationDetail] = Field([LocationDetail()],
        description="List of locations in which this Item appears.",
        min_items=1,
        # example provided by LocationDetail model
    )
    remarks: list[NonEmptyStr] = Field([],
        description="List of remarks and trivia about this item.",
        example=[
            "Activates the equipped Great Rune until death.",
            "Use animation is long and leaves you open to attacks.",
        ],
    )


import re
from enum import Enum
from operator import add
from itertools import repeat
from pathlib import Path
from typing import Any, overload
from pydantic.json import pydantic_encoder

from erdb.typing.params import ParamDict


def find_offset_indices(base_index: int, params: ParamDict, possible_maxima: list[int], increment: int = 1) -> tuple[map, range]:
    """
    Returns lists of valid indices from `base_index` value which offset
    until the highest possible maxima is reached.

    First list is possible indices and second is raw level integers.
    """
    def _find_offset_maxima() -> int | None:
        for maxima in sorted(possible_maxima, reverse=True): # from largest
            if (base_index + maxima * increment) in params.keys():
                return maxima
        return None

    maxima = _find_offset_maxima()
    assert maxima is not None

    levels = range(0, (maxima + 1) * increment, increment)
    return map(add, repeat(base_index), levels), levels

def prepare_writable_path(path: Path, default_filename: str) -> Path:
    (path if path.suffix == "" else path.parent).mkdir(parents=True, exist_ok=True)
    return path / default_filename if path.is_dir() else path

@overload
def remove_nulls(val: dict) -> dict: ...

@overload
def remove_nulls(val: list) -> list: ...

@overload
def remove_nulls(val: Any) -> Any: ...

def remove_nulls(val: dict | list | Any) -> dict | list | Any:
    """
    Recursively remove all None values from dictionaries and lists, and returns
    the result as a new dictionary or list.
    """
    if isinstance(val, list):
        return [remove_nulls(x) for x in val if x is not None]

    elif isinstance(val, dict):
        return {k: remove_nulls(v) for k, v in val.items() if v is not None}

    else:
        return val

def pydantic_encoder_no_nulls(obj: Any) -> Any:
    return remove_nulls(pydantic_encoder(obj))

def get_filename(name: str) -> str:
    return re.sub(r"(?u)[^-\w. &\[\]']", "", name)

def as_str(v: Any) -> str:
    return v.value if isinstance(v, Enum) else str(v)

def getattrstr(obj: Any, field: str) -> str:
    return as_str(getattr(obj, field))


import json
from math import floor
from pathlib import Path
from typing import Iterator, NamedTuple, Self


"""
Wrapper class for values, specifying base damage of the armament
and additional scaling from player attributes. Can be unpacked with:
    base, scaling = ValueType(1, 2)
"""
class ValueType(NamedTuple):
    base: float
    scaling: float

    @property
    def total(self) -> float:
        return floor(self.base + self.scaling)

    def __iter__(self) -> Iterator[tuple[float, float]]:
        return iter((self.base, self.scaling)) # type: ignore

class AttackPower(NamedTuple):
    physical: ValueType
    magic: ValueType
    fire: ValueType
    lightning: ValueType
    holy: ValueType

    def items(self):
        return zip(self._fields, self)

class StatusEffects(NamedTuple):
    bleed: ValueType
    frostbite: ValueType
    poison: ValueType
    scarlet_rot: ValueType
    sleep: ValueType
    madness: ValueType

    def items(self):
        return zip(self._fields, self)

class Attributes(NamedTuple):
    strength: int
    dexterity: int
    intelligence: int
    faith: int
    arcane: int

    def items(self):
        return zip(self._fields, self)

    @classmethod
    def from_string(cls, string: str) -> Self:
        parts = string.split(",")

        assert len(parts) == 5, "Invalid Attributes string"
        assert all(1 <= v <= 99 for v in map(int, parts)), "Invalid Attributes string"

        return cls(*map(int, parts))

    def __str__(self) -> str:
        return f"{self.strength},{self.dexterity},{self.intelligence},{self.faith},{self.arcane}"

class CorrectionAttack(NamedTuple):
    correction: dict[str, dict]
    override: dict[str, dict]
    ratio: dict[str, dict]

    @classmethod
    def from_dict(cls, data: dict) -> Self:
        return cls(data["correction"], data["override"], data["ratio"])

class CalculatorData(NamedTuple):
    armaments: dict[str, dict]
    reinforcements: dict[str, list[dict]]
    correction_attack: dict[str, dict[str, str]]
    correction_graph: dict[str, list[float]]

    @classmethod
    def create(cls, data_path: Path) -> Self:
        if not isinstance(data_path, Path):
            data_path = Path(data_path)

        with open(data_path / "armaments.json") as f:
            armaments = json.load(f)

        with open(data_path / "reinforcements.json") as f:
            reinforcements = json.load(f)

        with open(data_path / "correction-attack.json") as f:
            correction_attack = json.load(f)

        with open(data_path / "correction-graph.json") as f:
            correction_graph = json.load(f)

        return cls(armaments, reinforcements, correction_attack, correction_graph)

class ArmamentCalculator:
    _name: str
    _affinity: str
    _level: int

    # data cache relevant for this armament, affinity and level
    _affinity_properties: dict
    _requirements: dict[str, int]
    _reinforcement: dict
    _correction_attack: CorrectionAttack
    _correction_graph: dict[str, list[float]]

    def __init__(self, data: CalculatorData, name: str, affinity: str = "Standard", level: int = 0) -> None:
        self._name = name
        self._affinity = affinity
        self._level = level
        self._cache_data(data)

    def _cache_data(self, data: CalculatorData):
        """
        Cache some data we will be using for this particular armament/affinity/level combo,
        called every time the ArmamentCalculator instance is updated.
        """
        self._affinity_properties = data.armaments[self._name]["affinity"][self._affinity]
        reinforcement_id          = self._affinity_properties["reinforcement_id"]
        correction_attack_id      = self._affinity_properties["correction_attack_id"]

        self._requirements        = data.armaments[self._name]["requirements"]
        self._reinforcement       = data.reinforcements[str(reinforcement_id)][self._level]
        self._correction_attack   = CorrectionAttack.from_dict(data.correction_attack[str(correction_attack_id)])
        self._correction_graph    = data.correction_graph

    @property
    def name(self) -> str:
        return self._name

    @property
    def affinity(self) -> str:
        return self._affinity

    @property
    def level(self) -> int:
        return self._level

    def set_name(self, name: str, data: CalculatorData):
        self._name = name
        self._cache_data(data)

    def set_affinity(self, affinity: str, data: CalculatorData):
        self._affinity = affinity
        self._cache_data(data)

    def set_level(self, level: int, data: CalculatorData):
        self._level = level
        self._cache_data(data)

    """
    Calculate attack power of the weapon given player attributes.
    """
    def attack_power(self, attributes: Attributes) -> AttackPower:
        ret = {attack_type: self._get_base_and_scaled_damage(attack_type, attributes) for attack_type in AttackPower._fields}
        return AttackPower(**ret)

    """
    Retrieve base damage and scaled damage from the specific attack type and attributes.
    """
    def _get_base_and_scaled_damage(self, attack_type: str, attributes: Attributes) -> ValueType:
        base     = self._affinity_properties["damage"].get(attack_type, 0.0) * self._reinforcement["damage"][attack_type]
        scalings = [self._get_scaling_per_attribute(attack_type, attrib_name, attrib_value) for attrib_name, attrib_value in attributes.items()]
        low_cap  = min(scalings) # in case multiple attributes are not met, do not go below lowest scaling

        # return base damage and scaling which is the base * sum of scalings per every attribute
        return ValueType(base, base * max(low_cap, sum(scalings)))

    """
    Retrieve scaled damage for an attack type/player attribute/attribute value combo.
    """
    def _get_scaling_per_attribute(self, attack_type: str, attrib_name: str, attrib_value: int) -> float:
        # attack type does not scale with this attribute
        if not self._correction_attack.correction[attack_type][attrib_name]:
            return 0.0

        # get the impact ratio of the scaling for this attack type and attribute
        scaling_impact_ratio = self._correction_attack.ratio[attack_type][attrib_name]

        # requirement is not met, penalize scaling of this attack type for this attribute
        if attrib_value < self._requirements.get(attrib_name, 0):
            return 0.6 * (scaling_impact_ratio - 1) - 0.4

        # get scaling values for armament and its particular reinforcement level
        base_scaling  = self._affinity_properties["scaling"].get(attrib_name, 0.0)
        level_scaling = self._reinforcement["scaling"][attrib_name]

        # override base scaling if an override is defined
        base_scaling = self._correction_attack.override[attack_type].get(attrib_name, base_scaling)

        # get correction for scaling based on the attribute value
        correction_id      = self._affinity_properties["correction_calc_id"][attack_type]
        scaling_correction = self._correction_graph[str(correction_id)][attrib_value]

        # return actual scaled value for this attack type and attribute
        return scaling_impact_ratio - 1 + base_scaling * level_scaling * scaling_correction * scaling_impact_ratio

    """
    Calculate status effects of the weapon given player attributes.
    """
    def status_effects(self, attributes: Attributes) -> StatusEffects:
        ret = {effect_type: self._get_base_and_scaled_effect(effect_type, attributes) for effect_type in StatusEffects._fields}
        return StatusEffects(**ret)

    """
    Retrieve base and scaled status effect values from the specific effect type and attributes.
    """
    def _get_base_and_scaled_effect(self, effect_type: str, attributes: Attributes) -> ValueType:
        base     = self._affinity_properties["status_effects"].get(effect_type, 0.0)
        overlays = self._affinity_properties["status_effect_overlay"]
        level    = self._reinforcement["level"]

        # overwrite base value if the effect upgrades for the affinity
        if len(overlays) > level and effect_type in overlays[level]:
            base = overlays[level][effect_type]

        # retrieve scaling value if the effect can scale
        if correction_id := self._affinity_properties["correction_calc_id"].get(effect_type):
            base_scaling       = self._affinity_properties["scaling"].get("arcane", 0.0)
            level_scaling      = self._reinforcement["scaling"].get("arcane")
            scaling_correction = self._correction_graph[str(correction_id)][attributes.arcane]

            return ValueType(base, base * base_scaling * level_scaling * scaling_correction)
        
        return ValueType(base, 0.0)